from typing import List, Literal

from ....utils.class_registry import register_class

from .eula_3_8 import Eula_3_8 as E_3_8
from .eula_3_8 import IcetideVortex as IV_3_8
from .eula_3_8 import GlacialIllumination as GI_3_8
from .eula_3_8 import LightfallSword_3_8 as LS_3_8
from .eula_3_8 import FavoniusBladeworkEdel


class LightfallSword_3_5(LS_3_8):
    version: Literal["3.5"] = "3.5"
    damage: int = 2


class IcetideVortex(IV_3_8):
    version: Literal["3.5"] = "3.5"


class GlacialIllumination(GI_3_8):
    version: Literal["3.5"] = "3.5"


class Eula_3_5(E_3_8):
    version: Literal["3.5"]
    skills: List[FavoniusBladeworkEdel | IcetideVortex | GlacialIllumination] = []

    def _init_skills(self) -> None:
        self.skills = [FavoniusBladeworkEdel(), IcetideVortex(), GlacialIllumination()]


register_class(Eula_3_5 | LightfallSword_3_5)



"""
1. Lightning Rod will disappear when damage increase is triggered.
2. Talent card become a skill talent, and effect is same.
"""
from typing import Dict, List, Literal
from ....action import RemoveObjectAction
from ....event import AfterMakeDamageEventArguments, MakeDamageEventArguments
from ....match import Match
from ....patch.v43.characters.thunder_manifestation_4_3 import (
    GrievingEcho_4_3,
    LightningProbe,
    LightningRod_4_3,
    LightningStrikeProbe_4_3,
    StrifefulLightning,
    ThunderManifestation_4_3,
    ThunderingShackles,
)
from ....status.character_status.base import UsageCharacterStatus
from ....character.character_base import SkillTalent
from ....struct import Cost
from .....utils.class_registry import register_class
from ....character.character_base import ElementalNormalAttackBase
from ....consts import ELEMENT_TO_DAMAGE_TYPE, DieColor
from .....utils.desc_registry import DescDictType


class LightningStrikeProbe_4_4(LightningStrikeProbe_4_3):
    version: Literal["4.4"] = "4.4"


class LightningRod_4_4(LightningRod_4_3, UsageCharacterStatus):
    version: Literal["4.4"] = "4.4"

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        # do nor remove in make damage event, so talent can work
        return []

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Match
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class LightningProbe_4_4(LightningProbe):
    name: Literal["Lightning Probe"] = "Lightning Probe"
    version: Literal["4.4"] = "4.4"


class GrievingEcho_4_4(SkillTalent, GrievingEcho_4_3):
    version: Literal["4.4"] = "4.4"
    character_name: Literal["Thunder Manifestation"] = "Thunder Manifestation"
    skill: Literal["Strifeful Lightning"] = "Strifeful Lightning"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )


class ThunderManifestation_4_4(ThunderManifestation_4_3):
    version: Literal["4.4"] = "4.4"
    skills: List[
        ElementalNormalAttackBase
        | StrifefulLightning
        | ThunderingShackles
        | LightningProbe_4_4
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Thunderous Wingslash",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            StrifefulLightning(),
            ThunderingShackles(),
            LightningProbe_4_4(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Thunder Manifestation": {"descs": {"4.4": {"en-US": "", "zh-CN": ""}}},
    "SKILL_Thunder Manifestation_NORMAL_ATTACK/Thunderous Wingslash": {
        "descs": {
            "4.4": {"en-US": "Deals 1 Electro DMG.", "zh-CN": "造成1点雷元素伤害。"}
        }
    },
    "SKILL_Thunder Manifestation_ELEMENTAL_SKILL/Strifeful Lightning": {
        "descs": {
            "4.4": {
                "en-US": "Deals 3 Electro DMG to opposing characters affected by Lightning Rod. (If there are no eligible opposing characters, deals DMG to the active character instead)",  # noqa: E501
                "zh-CN": "对附属有雷鸣探知的敌方角色造成3点雷元素伤害。（如果敌方不存在符合条件角色，则改为对出战角色造成伤害）",  # noqa: E501
            }
        }
    },
    "CHARACTER_STATUS/Lightning Rod": {
        "descs": {
            "4.4": {
                "en-US": "While this status is active, can be triggered once and remove this status: DMG received by the attached character from Thunder Manifestation or its summons is increased by 1.\n(Only one of this status can exist on the field at once. Some of Thunder Manifestation's skills will target the character to which this is attached.)",  # noqa: E501
                "zh-CN": "此状态存在期间，可以触发1次并移除此状态：所附属角色受到雷音权现及其召唤物造成的伤害+1。\n（同一方场上最多存在一个此状态。雷音权现的部分技能，会以所附属角色为目标。）",  # noqa: E501
            }
        }
    },
    "SKILL_Thunder Manifestation_ELEMENTAL_BURST/Thundering Shackles": {
        "descs": {
            "4.4": {
                "en-US": "Deals 2 Electro DMG, summons 1 Thundering Shackles.",
                "zh-CN": "造成2点雷元素伤害，召唤轰雷禁锢。",
            }
        }
    },
    "SKILL_Thunder Manifestation_PASSIVE/Lightning Probe": {
        "descs": {
            "4.4": {
                "en-US": "(Passive) When battle begins, create a Lightning Strike Probe on the opponent's side of the field.",  # noqa: E501
                "zh-CN": "【被动】战斗开始时，在敌方场上生成雷霆探针。",
            }
        }
    },
    "TEAM_STATUS/Lightning Strike Probe": {
        "descs": {
            "4.4": {
                "en-US": "After a character on whose side of the field this card is on uses a Skill: Attach Lightning Rod to the active character on that side. (Once per Round)",  # noqa: E501
                "zh-CN": "所在阵营角色使用技能后：对所在阵营出战角色附属雷鸣探知。（每回合1次）",
            }
        },
    },
    "TALENT_Thunder Manifestation/Grieving Echo": {
        "descs": {
            "4.4": {
                "en-US": "Combat Action: When your active character is Thunder Manifestation, equip this card. After Thunder Manifestation equips this card, immediately use Strifeful Lightning. When Thunder Manifestation, with this card equipped, is in play, when opponents with Lightning Rod attached take DMG: You draw 1 card (Once per Round)\n(Your deck must contain Thunder Manifestation to add this card to your deck)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为雷音权现时，装备此牌。雷音权现装备此牌后，立刻使用一次雷墙倾轧。装备有此牌的雷音权现在场，附属有雷鸣探知的敌方角色受到伤害时：我方抓1张牌。（每回合1次）\n（牌组中包含雷音权现，才能加入牌组）",  # noqa: E501
            }
        },
    },
}


register_class(
    LightningRod_4_4
    | LightningStrikeProbe_4_4
    | GrievingEcho_4_4
    | ThunderManifestation_4_4,
    desc,
)



from typing import Literal, List
from pydantic import validator

from ....utils.class_registry import register_class
from ...consts import ElementType, FactionType, WeaponType, ELEMENT_TO_DAMAGE_TYPE
from .mob_1_0 import Mob_1_0
from ..character_base import (
    ElementalNormalAttackBase,
    ElementalSkillBase,
    ElementalBurstBase,
)


class MobMage_1_0(Mob_1_0):
    """
    Mob mages. They cannot carry weapons in default. Their normal
    attacks are 2 physical DMG, elemental skills are 3 element DMG,
    elemental bursts are 2 charges, 3 cost, and 5 element DMG.
    Their names should be BlablaMob, here Blabla should be one of elemental
    types.
    """

    name: Literal[
        "CryoMobMage",
        "HydroMobMage",
        "PyroMobMage",
        "ElectroMobMage",
        "GeoMobMage",
        "DendroMobMage",
        "AnemoMobMage",
    ]
    element: ElementType = ElementType.NONE
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | ElementalSkillBase | ElementalBurstBase
    ] = []
    faction: List[FactionType] = []
    weapon_type: WeaponType = WeaponType.CATALYST

    @validator("element")
    def element_fits_name(cls, v: ElementType, values, **kwargs):
        """
        Check if element type fits name.
        """
        if "name" not in values:
            raise AssertionError("Name not found.")
        type_in_name: str = values["name"][:-7].upper()
        element_type: str = v.value
        if type_in_name != element_type:
            raise ValueError(
                f"Element type {element_type} does not fit " f"name {type_in_name}."
            )
        return v

    def _init_skills(self):
        if self.element == ElementType.NONE:
            # set element by name
            element_name = self.name[:-7].upper()
            self.element = ElementType(element_name)
        element_name = self.element.value.lower()
        element_name = element_name[0].upper() + element_name[1:]
        normal_attack = ElementalNormalAttackBase(
            name=f"{element_name} Normal Attack",
            damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
            cost=ElementalNormalAttackBase.get_cost(self.element),
        )
        elemental_skill = ElementalSkillBase(
            name=f"{element_name} Elemental Skill",
            damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
            cost=ElementalSkillBase.get_cost(self.element),
        )
        elemental_burst = ElementalBurstBase(
            name=f"{element_name} Elemental Burst",
            damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
            cost=ElementalBurstBase.get_cost(self.element, 3, 2),
            damage=5,
        )
        self.skills = [normal_attack, elemental_skill, elemental_burst]


register_class(MobMage_1_0)



"""
With class registry, after version 4.3, new cards and characters do not need to
be defined in corresponding folders, and they will be defined in the patch.
"""


from ...utils import import_all_modules


import_all_modules(__file__, __name__)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DeclareRoundEndAttackSummonBase

from ...action import ActionTypes, Actions, RemoveObjectAction
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class SesshouSakura_3_7(DeclareRoundEndAttackSummonBase):
    name: Literal["Sesshou Sakura"] = "Sesshou Sakura"
    version: Literal["3.7"] = "3.7"
    usage: int = 3
    max_usage: int = 6
    damage_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    damage: int = 1
    renew_type: Literal["ADD"] = "ADD"
    extra_attack_usage: int = 4


# Skills


class YakanEvocationSesshouSakura(ElementalSkillBase):
    name: Literal["Yakan Evocation: Sesshou Sakura"] = "Yakan Evocation: Sesshou Sakura"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [
            self.charge_self(1),
            self.create_summon("Sesshou Sakura"),
        ]


class GreatSecretArtTenkoKenshin(ElementalBurstBase):
    name: Literal["Great Secret Art: Tenko Kenshin"] = "Great Secret Art: Tenko Kenshin"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret = super().get_actions(match)
        damage_action = ret[1]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        summons = match.player_tables[self.position.player_idx].summons
        sakura: SesshouSakura_3_7 | None = None
        for summon in summons:
            if summon.name == "Sesshou Sakura":
                sakura = summon
                break
        if sakura is not None:
            # destroy and add status
            ret = [RemoveObjectAction(object_position=sakura.position)] + ret
            ret.append(self.create_team_status("Tenko Thunderbolts"))
            if self.is_talent_equipped(match):
                # add status
                ret.append(self.create_character_status("The Shrine's Sacred Shade"))
        return ret


# Talents


class TheShrinesSacredShade_3_7(SkillTalent):
    name: Literal["The Shrine's Sacred Shade"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Yae Miko"] = "Yae Miko"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=2
    )
    skill: Literal[
        "Great Secret Art: Tenko Kenshin"
    ] = "Great Secret Art: Tenko Kenshin"


# character base


class YaeMiko_3_7(CharacterBase):
    name: Literal["Yae Miko"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase
        | YakanEvocationSesshouSakura
        | GreatSecretArtTenkoKenshin
    ] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="Spiritfox Sin-Eater",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            YakanEvocationSesshouSakura(),
            GreatSecretArtTenkoKenshin(),
        ]


register_class(YaeMiko_3_7 | TheShrinesSacredShade_3_7 | SesshouSakura_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class


from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import CharacterReviveEventArguments, SkillEndEventArguments

from ...action import Actions, CharacterReviveAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class HeraldOfFrost_4_0(AttackerSummonBase):
    name: Literal["Herald of Frost"] = "Herald of Frost"
    version: Literal["4.0"] = "4.0"
    usage: int = 3
    max_usage: int = 3
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO
    damage: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        if event.action.position.player_idx != self.position.player_idx:
            # not our player skill made damage, do nothing
            return []
        assert event.action.position.area == ObjectPositionType.SKILL
        characters = match.player_tables[self.position.player_idx].characters
        character = characters[event.action.position.character_idx]
        if character.name != "Qiqi":
            # not Qiqi, do nothing
            return []
        if event.action.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack, do nothing
            return []
        # select character with most damage taken
        selected_character = character
        for c in characters:
            if c.is_alive and c.damage_taken > selected_character.damage_taken:
                selected_character = c
        # heal character
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=selected_character.position,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ],
            )
        ]


# Skills


class AdeptusArtHeraldOfFrost(ElementalSkillBase):
    name: Literal["Adeptus Art: Herald of Frost"] = "Adeptus Art: Herald of Frost"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        return [
            self.charge_self(1),
            self.create_summon("Herald of Frost"),
        ]


class AdeptusArtPreserverOfFortune(ElementalBurstBase):
    name: Literal[
        "Adeptus Art: Preserver of Fortune"
    ] = "Adeptus Art: Preserver of Fortune"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=3
    )

    revive_usage: int = 2

    def event_handler_CHARACTER_REVIVE(
        self, event: CharacterReviveEventArguments, match: Any
    ) -> List[Actions]:
        """
        When self is revived, reset revive usage
        """
        if (
            event.action.player_idx == self.position.player_idx
            and event.action.character_idx == self.position.character_idx
        ):
            self.revive_usage = 2
        return []

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Make damage and create status. Then if has talent and has revive usage,
        revive all defeated characters with 2 hp.
        """
        ret = super().get_actions(match) + [
            self.create_team_status("Fortune-Preserving Talisman")
        ]
        if not self.is_talent_equipped(match):
            # no talent
            return ret
        # has talent
        if self.revive_usage <= 0:
            # no revive usage
            return ret
        defeated_characters: List[CharacterBase] = []
        characters = match.player_tables[self.position.player_idx].characters
        for c in characters:
            if c.is_defeated:
                defeated_characters.append(c)
        if len(defeated_characters) == 0:
            # no defeated character
            return ret
        # has defeated character, revive them
        self.revive_usage -= 1
        for c in defeated_characters:
            ret.append(
                CharacterReviveAction(
                    player_idx=self.position.player_idx,
                    character_idx=c.position.character_idx,
                    revive_hp=2,
                )
            )
        return ret


# Talents


class RiteOfResurrection_4_0(SkillTalent):
    name: Literal["Rite of Resurrection"]
    version: Literal["4.0"] = "4.0"
    character_name: Literal["Qiqi"] = "Qiqi"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=5, charge=3
    )
    skill: Literal[
        "Adeptus Art: Preserver of Fortune"
    ] = "Adeptus Art: Preserver of Fortune"


# character base


class Qiqi_4_0(CharacterBase):
    name: Literal["Qiqi"]
    version: Literal["4.0"] = "4.0"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase
        | AdeptusArtHeraldOfFrost
        | AdeptusArtPreserverOfFortune
    ] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Ancient Sword Art",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            AdeptusArtHeraldOfFrost(),
            AdeptusArtPreserverOfFortune(),
        ]


register_class(Qiqi_4_0 | RiteOfResurrection_4_0 | HeraldOfFrost_4_0)



import shlex
from typing import Any, List

from .consts import ObjectPositionType, ObjectType


def _query(object_position: Any, match: Any, command: List[str]) -> List[Any]:
    """
    Function to deal with a query. `and` should not contain in command when calling,
    split it before calling this function.
    """
    assert len(command) > 0, "command is empty"
    current_objs = []
    if command[0] == "self":
        assert object_position.area in [
            ObjectPositionType.CHARACTER,
            ObjectPositionType.CHARACTER_STATUS,
            ObjectPositionType.SKILL,
        ], "self can only be used for objects in character skill or character_status"
        current_objs = [
            match.player_tables[object_position.player_idx].characters[
                object_position.character_idx
            ]
        ]
    elif command[0] == "both":
        current_objs = [match.player_tables[0], match.player_tables[1]]
    elif command[0] == "our":
        current_objs = [match.player_tables[object_position.player_idx]]
    elif command[0] == "opponent":
        current_objs = [match.player_tables[1 - object_position.player_idx]]
    else:
        raise ValueError(
            f"first token must be both, our, opponent, or self. input is {command[0]}"
        )
    for cmd in command[1:]:
        if cmd in [
            "active",
            "prev",
            "next",
            "deck",
            "hand",
            "character",
            "team_status",
            "summon",
            "support",
        ]:
            for obj in current_objs:
                assert obj.__class__.__name__ == "PlayerTable", (
                    f"object must be a player_table with command {cmd}, "
                    f"current object is {obj.__class__.__name__}"
                )
            if cmd == "active":
                current_objs = [x.get_active_character() for x in current_objs]
            elif cmd == "prev":
                current_objs = [
                    x.characters[x.previous_character_idx()] for x in current_objs
                ]
                current_objs = [x for x in current_objs if x is not None]
            elif cmd == "next":
                current_objs = [
                    x.characters[x.next_character_idx()] for x in current_objs
                ]
                current_objs = [x for x in current_objs if x is not None]
            elif cmd == "deck":
                current_objs = sum([x.table_deck for x in current_objs], [])
            elif cmd == "hand":
                current_objs = sum([x.hands for x in current_objs], [])
            elif cmd == "character":
                current_objs = sum([x.characters for x in current_objs], [])
            elif cmd == "team_status":
                current_objs = sum([x.team_status for x in current_objs], [])
            elif cmd == "summon":
                current_objs = sum([x.summons for x in current_objs], [])
            else:
                assert cmd == "support"
                current_objs = sum([x.supports for x in current_objs], [])
        elif cmd in ["weapon", "artifact", "talent", "status", "skill"]:
            for obj in current_objs:
                assert (
                    obj.__class__.__name__ != "PlayerTable"
                    and obj.type == ObjectType.CHARACTER
                ), (
                    f"object must be a character with command {cmd}, "
                    f"current object is {obj.__class__.__name__}"
                )
            if cmd in ["weapon", "artifact", "talent"]:
                current_objs = [getattr(x, cmd) for x in current_objs]
                current_objs = [x for x in current_objs if x is not None]
            elif cmd == "status":
                current_objs = sum([x.status for x in current_objs], [])
            else:
                assert cmd == "skill"
                current_objs = sum([x.skills for x in current_objs], [])
        elif "=" in cmd:
            key, value = cmd.split("=")
            current_objs = [
                x
                for x in current_objs
                if str(getattr(x, key, None)).lower() == value.lower()
            ]
        else:
            raise ValueError(f"unknown command {cmd}")
    return current_objs


def query(object_position: Any, match: Any, command: str) -> List[Any]:
    """
    Receives object position, match, and get objects from match based on command.

    Command contains:
    - `both / our / opponent` to select player_table, or `self` to select this character
        (caller area must be character or character_status). must appear first.
    - `[active / prev / next]` or
        `[deck / hand / character / team_status / summon / support]`
        for player_table, `[weapon / artifact / talent]` or `[skill / status]` for
        character, to select an object (first group) or object lists (second group).
        Note with `both`, select one object may also receive two objects.
    - `['key=value']` to filter current objects that `x.key == value`. All types will be
        converted to str to compare. it can be used multiple times. command can be
        surrounded with quotes, and if not space in key and value, quotes can be
        omitted.
    - `and` to make multiple queries, and their results are combined.

    Samples:
    - `self` to get self character when called by a character status
    - `opponent active status 'name=Seed of Skandha'` to get status
    - `opponent character status name=Refraction` to get status from all opponent
        characters
    - `both summon` to get all summons on field
    - `self and our active and our next` to select characters
    """
    tokens = shlex.split(command)
    results = []
    while True:
        and_index = -1
        if "and" in tokens:
            and_index = tokens.index("and")
        if and_index == -1:
            results.extend(_query(object_position, match, tokens))
            break
        else:
            results.extend(_query(object_position, match, tokens[:and_index]))
            tokens = tokens[and_index + 1 :]
    return results


def query_one(
    object: Any, match: Any, command: str, allow_multiple: bool = False
) -> Any | None:
    """
    Receives object position, match, and get one object from match based on command.
    If no object is found, return None. For command structure, see `query` function.
    """
    result = query(object, match, command)
    if len(result) == 0:
        return None
    if not allow_multiple and len(result) > 1:
        raise ValueError(f"Multiple objects found for command {command}")
    return result[0]


def _satisfy_single_position(
    object_position: Any, command: List[str], match: Any | None
) -> bool:
    """
    deal with one object position and command, return if the object position satisfies
    the command. match is optional. if not provided, some command will raise error.
    """
    for cmd in command:
        if "=" not in cmd:
            raise ValueError(f"command {cmd} is not valid")
        key, value = cmd.split("=")
        if key == "pidx":
            if object_position.player_idx != int(value):
                return False
        elif key == "cidx":
            if object_position.character_idx != int(value):
                return False
        elif key == "area":
            if object_position.area.name.lower() != value.lower():
                return False
        elif key == "id":
            if object_position.id != int(value):
                return False
        elif key == "active":
            value = value.lower()
            if object_position.character_idx == -1:
                raise ValueError("active can only be used with valid character index")
            if match is None:
                raise AssertionError(
                    "match is not provided when command contains active"
                )
            if value not in ["true", "false"]:
                raise ValueError(
                    f"active value {value} is not valid, should be true or false"
                )
            table = match.player_tables[object_position.player_idx]
            active_idx = table.active_character_idx
            value = value == "true"
            if (active_idx == object_position.character_idx) != value:
                return False
        else:
            raise ValueError(f"unknown command {cmd}")
    return True


def _satisfy_between_position(source: Any, target: Any, command: List[str]) -> bool:
    """
    deal with two object positions and command, return if the two object positions
    satisfy the command.
    """
    for cmd in command:
        if "=" not in cmd:
            raise ValueError(f"command {cmd} is not valid")
        key, value = cmd.split("=")
        if value not in ["same", "diff"]:
            raise ValueError(f"value {value} is not valid, should be same or diff")
        value = value == "same"
        if key == "pidx":
            if (source.player_idx == target.player_idx) != value:
                return False
        elif key == "cidx":
            if (source.character_idx == target.character_idx) != value:
                return False
        elif key == "area":
            if (source.area == target.area) != value:
                return False
        elif key == "id":
            if (source.id == target.id) != value:
                return False
        else:
            raise ValueError(f"unknown command {cmd}")
    return True


def _satisfy_one(
    source: Any, target: Any | None, command: List[str], match: Any | None
) -> bool:
    """
    deal with one command without and.
    """
    assert command != "and", "command should not contain and"
    if command[0] == "both":
        assert target is not None, "target is not provided"
        return _satisfy_between_position(source, target, command[1:])
    elif command[0] == "source":
        return _satisfy_single_position(source, command[1:], match)
    elif command[0] == "target":
        assert target is not None, "target is not provided"
        return _satisfy_single_position(target, command[1:], match)
    else:
        raise ValueError(f"unknown command {command[0]}")


def satisfy(
    source: Any, command: str, target: Any | None = None, match: Any | None = None
) -> bool:
    """
    receives source `ObjectPosition`, target `ObjectPosition` and match. target and
    match are optional. without them, related command string will raise error.

    - `source / target` to select one object position, or `both` to compare between two
        positions
    - if one position selected: `[pidx=? / cidx=? / area=? / active=(true|false)]` to
        check if position fulfills the situation. for area names, they are case
        insensitive. can use multiple times, and all of them should pass.
    - if two position selected (i.e. `both`), `[(pidx|cidx|area|id)=(same|diff)]` to
        compare two positions are same or not. can use multiple times, and all of them
        should pass.
    - `and` to make multiple checks. All check should pass.

    samples:
    - `source area=support` is placed to support
    - `source area=character and target area=skill and both pidx=same cidx=same` is
        equipped to a character, and target is skill of this character
    - `both pidx=same and source area=hand and target area=skill` self in hand and
        target is this player use skill
    """
    tokens = shlex.split(command)
    while True:
        and_index = -1
        if "and" in tokens:
            and_index = tokens.index("and")
        if and_index == -1:
            return _satisfy_one(source, target, tokens, match)
        else:
            if not _satisfy_one(source, target, tokens[:and_index], match):
                return False
            tokens = tokens[and_index + 1 :]



from typing import Any, Dict, List, Literal

from ....utils.class_registry import register_class

from ...event import SkillEndEventArguments

from ...action import (
    Actions,
    CreateObjectAction,
    RemoveCardAction,
    SkillEndAction,
    SwitchCharacterAction,
    UseSkillAction,
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    AOESkillBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillBase,
    SkillTalent,
)


# Skills


class StellarRestoration(ElementalSkillBase):
    name: Literal["Stellar Restoration"] = "Stellar Restoration"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        If Lightning Stiletto in hand, do not create another one.
        """
        table = match.player_tables[self.position.player_idx]
        hands = table.hands[:]
        if table.using_hand is not None:
            hands.append(table.using_hand)
        found_stiletto = False
        for card in hands:
            if card.name == "Lightning Stiletto":
                found_stiletto = True
                break
        if not found_stiletto and len(hands) < match.config.max_hand_size:
            # not found stiletto and hand not full
            position = ObjectPosition(
                player_idx=self.position.player_idx, area=ObjectPositionType.HAND, id=-1
            )
            return super().get_actions(match) + [
                CreateObjectAction(
                    object_name="Lightning Stiletto",
                    object_position=position,
                    object_arguments={},
                )
            ]
        return super().get_actions(match)


class StarwardSword(ElementalBurstBase, AOESkillBase):
    name: Literal["Starward Sword"] = "Starward Sword"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    back_damage: int = 3
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=4,
        charge=3,
    )


# Talents


class ThunderingPenance_3_3(SkillTalent):
    name: Literal["Thundering Penance"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Keqing"] = "Keqing"
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3)
    skill: Literal["Stellar Restoration"] = "Stellar Restoration"


class LightningStiletto_3_3(SkillTalent):
    """
    From tests, Lightning Stiletto performs almost same of skill talent,
    except it will not equip the card to the character.
    (e.g. Vermillion Hereafter can decrease its cost, although it claimed that
    only can decrease the cost of talent or normal attack.)
    """

    name: Literal["Lightning Stiletto"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Keqing"] = "Keqing"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO,
        elemental_dice_number=3,
    )
    skill: Literal["Stellar Restoration"] = "Stellar Restoration"
    newly_created: bool = True
    use_card: bool = False

    def get_character_position(self, match: Any) -> ObjectPosition | None:
        """
        Find alive character index. If not found, return -1
        """
        characters = match.player_tables[self.position.player_idx].characters
        for character in characters:
            if (
                character.name == self.character_name
                and character.is_alive
                and not character.is_stunned
            ):
                return character.position
        return None

    def is_valid(self, match: Any) -> bool:
        return self.get_character_position(match) is not None

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        pos = self.get_character_position(match)
        assert pos is not None
        return [pos]

    def get_actions(self, target: ObjectPosition | None, match: Any) -> List[Actions]:
        """
        find keqing position, switch to it and use elemental skill immediately.
        No need to remove this card, as it will be removed after skill done.
        """
        assert target is not None
        # update skill position
        skills: List[SkillBase] = match.get_object(target).skills
        for skill in skills:
            if skill.name == self.skill:
                break
        else:
            raise AssertionError("Skill not found")
        ret: List[Actions] = []
        # if not active, switch to it
        active_idx = match.player_tables[self.position.player_idx].active_character_idx
        if active_idx != target.character_idx:
            ret.append(
                SwitchCharacterAction(
                    player_idx=self.position.player_idx,
                    character_idx=target.character_idx,
                )
            )
        # use skill
        ret.append(
            UseSkillAction(
                skill_position=skill.position,
            )
        )
        # skill used, add SkillEndAction
        ret.append(
            SkillEndAction(
                position=skill.position,
                target_position=match.player_tables[1 - skill.position.player_idx]
                .get_active_character()
                .position,
                skill_type=skill.skill_type,
            )
        )
        self.use_card = True
        return ret

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        If our Keqing use elemental skill, remove this card and create electro
        infusion status for her.
        """
        ret: List[Actions] = []
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            source_area=ObjectPositionType.HAND,
            target_area=ObjectPositionType.SKILL,
        ):
            # this card not in hand, or not our character use skill
            return []
        skill: SkillBase = match.get_object(event.action.position)
        if skill.name != self.skill:
            # not elemental skill
            return []
        if self.newly_created:
            # this card is newly created, mark False and do nothing.
            self.newly_created = False
            return []
        # remove this card
        table = match.player_tables[self.position.player_idx]
        hands = table.hands[:]
        if table.using_hand is not None:
            hands.append(table.using_hand)
        for i, card in enumerate(hands):
            if card.id == self.id:
                ret.append(
                    RemoveCardAction(
                        position=self.position,
                        remove_type="USED" if self.use_card else "BURNED",
                    )
                )
                break
        else:
            raise AssertionError("Card not in hand")
        # generate args
        args: Dict = {"mark": "Keqing"}
        character = match.player_tables[skill.position.player_idx].characters[
            skill.position.character_idx
        ]
        if character.talent is not None:
            args["usage"] = 3
            args["max_usage"] = 3
            args["talent_activated"] = True
        # create electro infusion status
        ret.append(
            CreateObjectAction(
                object_name="Electro Elemental Infusion",
                object_position=character.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments=args,
            )
        )
        return ret


# character base


class Keqing_3_3(CharacterBase):
    name: Literal["Keqing"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[PhysicalNormalAttackBase | StellarRestoration | StarwardSword] = []
    faction: List[FactionType] = [
        FactionType.LIYUE,
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Yunlai Swordsmanship",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            StellarRestoration(),
            StarwardSword(),
        ]


register_class(Keqing_3_3 | ThunderingPenance_3_3 | LightningStiletto_3_3)



from typing import Literal

from .....utils.class_registry import register_class

from ....consts import WeaponType
from .base import WeaponBase
from ....struct import Cost


class VanillaWeapon_3_3(WeaponBase):
    name: Literal[
        "Magic Guide",
        "Raven Bow",
        "Traveler's Handy Sword",
        "White Iron Greatsword",
        "White Tassel",
    ]
    version: Literal["3.3"] = "3.3"
    weapon_type: WeaponType = WeaponType.OTHER

    cost: Cost = Cost(same_dice_number=2)

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == "Magic Guide":
            self.weapon_type = WeaponType.CATALYST
        elif self.name == "Raven Bow":
            self.weapon_type = WeaponType.BOW
        elif self.name == "Traveler's Handy Sword":
            self.weapon_type = WeaponType.SWORD
        elif self.name == "White Iron Greatsword":
            self.weapon_type = WeaponType.CLAYMORE
        else:
            assert self.name == "White Tassel"
            self.weapon_type = WeaponType.POLEARM


register_class(VanillaWeapon_3_3)



from typing import List, Literal

from ....utils.class_registry import register_class
from .yanfei_3_8 import (
    DoneDeal,
    SealOfApproval,
    Yanfei_3_8,
    RightOfFinalInterpretation_3_8,
    SignedEdict as SE_3_8,
)


class SignedEdict(SE_3_8):
    max_usage: int = 2
    version: Literal["4.2"] = "4.2"


class RightOfFinalInterpretation_4_2(RightOfFinalInterpretation_3_8):
    """
    Draw card action is performed by Scarlet Seal.
    """

    version: Literal["4.2"] = "4.2"
    draw_card: bool = True


class Yanfei_4_2(Yanfei_3_8):
    version: Literal["4.2"] = "4.2"
    skills: List[SealOfApproval | SignedEdict | DoneDeal] = []

    def _init_skills(self) -> None:
        self.skills = [SealOfApproval(), SignedEdict(), DoneDeal()]


register_class(Yanfei_4_2 | RightOfFinalInterpretation_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_base_class, register_class

from ...action import (
    ActionTypes,
    Actions,
    CharacterReviveAction,
    CreateObjectAction,
    MakeDamageAction,
)
from ...struct import Cost, ObjectPosition
from ...modifiable_values import DamageValue
from ...consts import CostLabels, DamageElementalType, DamageType, ObjectPositionType
from ...object_base import EventCardBase


class FoodCardBase(EventCardBase):
    """
    Base class for food cards.
    """

    cost_label: int = (
        CostLabels.CARD.value | CostLabels.FOOD.value | CostLabels.EVENT.value
    )
    can_eat_only_if_damaged: bool

    def eat_target(self, match: Any) -> List[int]:
        """
        Get target character idxs that can eat.
        """
        ret: List[int] = []
        table = match.player_tables[self.position.player_idx]
        for idx, character in enumerate(table.characters):
            if character.is_defeated:
                continue
            is_full = False
            for status in character.status:
                if status.name == "Satiated":
                    is_full = True
                    break
            if is_full:
                continue
            if self.can_eat_only_if_damaged and character.damage_taken <= 0:
                continue
            ret.append(idx)
        return ret

    def is_valid(self, match: Any) -> bool:
        """
        If there is at least one target that can eat.
        """
        return len(self.eat_target(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Get targets of this card.
        """
        ret: List[ObjectPosition] = []
        characters = match.player_tables[self.position.player_idx].characters
        for idx in self.eat_target(match):
            ret.append(characters[idx].position)
        return ret

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        """
        Add Satiated status to target character.
        """
        assert target is not None
        pos = target.set_area(ObjectPositionType.CHARACTER_STATUS)
        return [
            CreateObjectAction(
                object_name="Satiated", object_position=pos, object_arguments={}
            )
        ]


register_base_class(FoodCardBase)


class AllCharacterFoodCard(FoodCardBase):
    """
    Food card that can be eaten by all characters.
    """

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        As it fulfills all characters, no need to specify target.
        """
        return []

    def get_actions(self, target: ObjectPosition | None, match: Any) -> List[Actions]:
        """
        It will add status to all characters that is not satiated.
        """
        ret: List[Actions] = []
        targets = self.eat_target(match)
        assert len(targets) > 0
        characters = match.player_tables[self.position.player_idx].characters
        for t in targets:
            character = characters[t]
            ret += self.one_character_action(character.position)
        return ret

    def one_character_action(self, pos: ObjectPosition) -> List[Actions]:
        """
        Actions that will be applied to one character. Default is to add
        Satiated and status that has same name as the card.
        """
        ret: List[Actions] = []
        pos = pos.set_area(ObjectPositionType.CHARACTER_STATUS)
        ret.append(
            CreateObjectAction(
                object_name="Satiated", object_position=pos, object_arguments={}
            )
        )
        ret.append(
            CreateObjectAction(
                object_name=self.name, object_position=pos, object_arguments={}
            )
        )
        return ret


class JueyunGuoba_3_3(FoodCardBase):
    name: Literal["Jueyun Guoba"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=ret[0].object_position,
                object_arguments={},
            )
        )
        return ret


class AdeptusTemptation_3_3(FoodCardBase):
    name: Literal["Adeptus' Temptation"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(any_dice_number=2)

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=ret[0].object_position,
                object_arguments={},
            )
        )
        return ret


class LotusFlowerCrisp_3_3(FoodCardBase):
    name: Literal["Lotus Flower Crisp"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=ret[0].object_position,
                object_arguments={},
            )
        )
        return ret


class NorthernSmokedChicken_3_3(FoodCardBase):
    name: Literal["Northern Smoked Chicken"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=ret[0].object_position,
                object_arguments={},
            )
        )
        return ret


class SweetMadame_3_3(FoodCardBase):
    name: Literal["Sweet Madame"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost()

    can_eat_only_if_damaged: bool = True

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | MakeDamageAction]:
        ret: List[CreateObjectAction | MakeDamageAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        ret.append(
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=target,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                ],
            )
        )
        return ret


class MondstadtHashBrown_3_3(FoodCardBase):
    name: Literal["Mondstadt Hash Brown"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)

    can_eat_only_if_damaged: bool = True

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | MakeDamageAction]:
        ret: List[CreateObjectAction | MakeDamageAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        ret.append(
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=target,
                        damage=-2,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                ],
            )
        )
        return ret


class MushroomPizza_3_3(FoodCardBase):
    name: Literal["Mushroom Pizza"]
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(same_dice_number=1)

    can_eat_only_if_damaged: bool = True

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | MakeDamageAction]:
        ret: List[CreateObjectAction | MakeDamageAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        ret.append(
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=target,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=self.cost.copy(),
                    )
                ],
            )
        )
        create_1 = ret[0]
        assert create_1.type == ActionTypes.CREATE_OBJECT
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=create_1.object_position,
                object_arguments={},
            )
        )
        return ret


class MintyMeatRolls_3_4(FoodCardBase):
    name: Literal["Minty Meat Rolls"]
    version: Literal["3.4"] = "3.4"
    cost: Cost = Cost(same_dice_number=1)

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=ret[0].object_position,
                object_arguments={"version": self.version},
            )
        )
        return ret


class MintyMeatRolls_3_3(MintyMeatRolls_3_4):
    version: Literal["3.3"]


class TeyvatFriedEgg_4_1(FoodCardBase):
    name: Literal["Teyvat Fried Egg"]
    version: Literal["4.1"] = "4.1"
    cost: Cost = Cost(same_dice_number=2)

    can_eat_only_if_damaged: bool = False

    def is_valid(self, match: Any) -> bool:
        """
        Has target, and not used this round.
        """
        for s in match.player_tables[self.position.player_idx].team_status:
            if s.name == "Revive on cooldown":
                return False
        return len(self.get_targets(match)) > 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        Can only target defeated characters
        """
        return [
            character.position
            for character in match.player_tables[self.position.player_idx].characters
            if character.is_defeated
        ]

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction | CharacterReviveAction]:
        ret: List[CreateObjectAction | CharacterReviveAction] = []
        ret = list(super().get_actions(target, match))
        assert len(ret) == 1
        assert target is not None
        create_1 = ret[0]
        assert create_1.type == ActionTypes.CREATE_OBJECT
        ret.append(
            CreateObjectAction(
                object_name="Revive on cooldown",
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        )
        ret = [
            CharacterReviveAction(
                player_idx=target.player_idx,
                character_idx=target.character_idx,
                revive_hp=1,
            )
        ] + ret
        return ret


class TeyvatFriedEgg_3_7(TeyvatFriedEgg_4_1):
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=3)


class SashimiPlatter_3_7(FoodCardBase):
    name: Literal["Sashimi Platter"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(same_dice_number=1)

    can_eat_only_if_damaged: bool = False

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[CreateObjectAction]:
        ret = super().get_actions(target, match)
        assert len(ret) == 1
        ret.append(
            CreateObjectAction(
                object_name=self.name,
                object_position=ret[0].object_position,
                object_arguments={},
            )
        )
        return ret


class TandooriRoastChicken_3_7(AllCharacterFoodCard):
    name: Literal["Tandoori Roast Chicken"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(any_dice_number=2)

    can_eat_only_if_damaged: bool = False


class ButterCrab_3_7(AllCharacterFoodCard):
    name: Literal["Butter Crab"]
    version: Literal["3.7"] = "3.7"
    cost: Cost = Cost(any_dice_number=2)

    can_eat_only_if_damaged: bool = False


register_class(
    JueyunGuoba_3_3
    | AdeptusTemptation_3_3
    | LotusFlowerCrisp_3_3
    | NorthernSmokedChicken_3_3
    | SweetMadame_3_3
    | MondstadtHashBrown_3_3
    | MushroomPizza_3_3
    | MintyMeatRolls_3_4
    | MintyMeatRolls_3_3
    | TeyvatFriedEgg_4_1
    | TeyvatFriedEgg_3_7
    | SashimiPlatter_3_7
    | TandooriRoastChicken_3_7
    | ButterCrab_3_7
)



from typing import List, Literal

from ..character_base import PhysicalNormalAttackBase

from ....utils.class_registry import register_class

from ...consts import DieColor
from ...struct import Cost
from .razor_3_8 import Awakening_4_2, Razor_3_8 as R_3_8
from .razor_3_8 import LightningFang as LF_3_8
from .razor_3_8 import ClawAndThunder


class LightningFang(LF_3_8):
    damage: int = 5
    cost: Cost = Cost(
        elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=3, charge=3
    )


class Awakening_3_3(Awakening_4_2):
    version: Literal["3.3"] = "3.3"
    cost: Cost = Cost(elemental_dice_color=DieColor.ELECTRO, elemental_dice_number=4)
    usage: int = 999
    max_usage: int = 999


class Razor_3_3(R_3_8):
    version: Literal["3.3"]
    skills: List[PhysicalNormalAttackBase | ClawAndThunder | LightningFang] = []
    max_charge: int = 3

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Steel Fang",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ClawAndThunder(),
            LightningFang(),
        ]


register_class(Razor_3_3 | Awakening_3_3)



from typing import Dict, Literal
import pytest
from lpsim.utils.desc_registry import DescDictType

from lpsim.server.status.team_status.base import (
    ElementalInfusionTeamStatus,
    TeamStatusBase,
)
from lpsim.utils.class_registry import get_instance, register_class

from lpsim.agents.random_agent import RandomAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.action import CreateDiceAction, DrawCardAction, MoveObjectAction
from lpsim.server.consts import DamageElementalType, ObjectPositionType
from lpsim.server.deck import Deck
from lpsim.server.interaction import SwitchCardResponse
from lpsim.server.object_base import ObjectBase
from lpsim.server.struct import ObjectPosition
from lpsim.server.match import Match, MatchConfig


def test_object_position_validation():
    p1 = ObjectPosition(
        player_idx=0,
        character_idx=1,
        area=ObjectPositionType.CHARACTER_STATUS,
        id=345,
    )
    p2 = ObjectPosition(
        player_idx=1, character_idx=0, area=ObjectPositionType.CHARACTER, id=123
    )
    match = Match()
    match.player_tables[0].active_character_idx = 1
    match.player_tables[1].active_character_idx = 0
    assert p1.check_position_valid(p2, match)
    assert not p1.check_position_valid(p2, match, player_idx_same=True)
    assert not p1.check_position_valid(p2, match, character_idx_same=True)
    assert not p1.check_position_valid(
        p2, match, player_idx_same=True, character_idx_same=True
    )
    assert not p1.check_position_valid(p2, match, area_same=True)
    assert not p1.check_position_valid(p2, match, id_same=True)
    assert p1.check_position_valid(p2, match, area_same=False)
    assert p1.check_position_valid(p2, match, id_same=False)
    assert p1.check_position_valid(
        p2, match, source_area=ObjectPositionType.CHARACTER_STATUS
    )
    assert p1.check_position_valid(p2, match, target_area=ObjectPositionType.CHARACTER)
    assert p1.check_position_valid(p2, match, source_is_active_character=True)
    assert p1.check_position_valid(p2, match, target_is_active_character=True)
    assert not p1.check_position_valid(
        p2, match, source_area=ObjectPositionType.CHARACTER
    )
    assert not p1.check_position_valid(
        p2, match, target_area=ObjectPositionType.CHARACTER_STATUS
    )
    assert not p1.check_position_valid(p2, match, source_is_active_character=False)
    assert not p1.check_position_valid(p2, match, target_is_active_character=False)
    assert p1.check_position_valid(p2, match, player_idx_same=False)
    assert p1.check_position_valid(
        p2, match, player_idx_same=False, character_idx_same=False
    )
    assert not p1.check_position_valid(
        p2, match, player_idx_same=True, character_idx_same=False
    )
    assert not p1.check_position_valid(
        p2, match, player_idx_same=False, character_idx_same=True
    )


def test_objectbase_wrong_handler_name():
    class WrongEventHandler(ObjectBase):
        name: Literal["WrongEventHandler"] = "WrongEventHandler"
        position: ObjectPosition = ObjectPosition(
            player_idx=-1,
            area=ObjectPositionType.INVALID,
            id=-1,
        )

        def event_handler_NOT_EXIST(self, event):
            ...  # pragma: no cover

    with pytest.raises(ValueError):
        _ = WrongEventHandler()

    class WrongValueModifier(ObjectBase):
        name: Literal["WrongValueModifier"] = "WrongValueModifier"
        position: ObjectPosition = ObjectPosition(
            player_idx=-1,
            area=ObjectPositionType.INVALID,
            id=-1,
        )

        def value_modifier_NOT_EXIST(self, value, match, mode):
            ...  # pragma: no cover

    with pytest.raises(ValueError):
        _ = WrongValueModifier()


def test_match_config_and_match_errors():
    config = MatchConfig()
    assert config.check_config()
    config.initial_hand_size = 11
    assert not config.check_config()
    config.initial_hand_size = 5
    config.initial_card_draw = 11
    assert not config.check_config()
    config.max_hand_size = 999
    config.card_number = 10
    assert not config.check_config()
    config.initial_card_draw = 2
    config.initial_dice_number = 20
    assert not config.check_config()
    config.initial_dice_number = 8
    config = MatchConfig(initial_hand_size=-1)
    assert not config.check_config()
    config = MatchConfig(initial_card_draw=-1)
    assert not config.check_config()
    config = MatchConfig(initial_dice_number=-1)
    assert not config.check_config()
    config = MatchConfig(initial_dice_reroll_times=-1)
    assert not config.check_config()
    config = MatchConfig(card_number=-1)
    assert not config.check_config()
    config = MatchConfig(max_same_card_number=-1)
    assert not config.check_config()
    config = MatchConfig(character_number=-1)
    assert not config.check_config()
    config = MatchConfig(max_round_number=0)
    assert not config.check_config()
    config = MatchConfig(max_round_number=-1)
    assert not config.check_config()
    config = MatchConfig(max_hand_size=-1)
    assert not config.check_config()
    config = MatchConfig(max_dice_number=-1)
    assert not config.check_config()
    config = MatchConfig(max_summon_number=-1)
    assert not config.check_config()
    config = MatchConfig(max_support_number=-1)
    assert not config.check_config()
    config = MatchConfig(history_level=-1)
    assert not config.check_config()
    match = Match()
    match.config = config
    assert not match.start()[0]
    match = Match()
    match.player_tables = []
    assert not match.start()[0]
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:GeoMob*3
        Strategize*100
        """
    )
    with pytest.raises(AssertionError):
        match.set_deck([])
    match.set_deck([deck, deck])
    assert not match.start()[0]
    deck = Deck.from_str(
        """
        default_version:4.0
        character:GeoMob*3
        Strategize*30
        """
    )
    match = Match()
    match.config.max_same_card_number = 30
    match.set_deck([deck, deck])
    assert match.start()[0]
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = RandomAgent(player_idx=1)
    non_exist_tested = False
    respond_no_request = False
    req = None
    resp = None
    for _ in range(30):
        if len(match.requests) == 0 and not respond_no_request and non_exist_tested:
            assert not match.need_respond(0)
            assert not match.need_respond(1)
            assert resp is not None
            with pytest.raises(ValueError):
                match.respond(resp)
        if match.need_respond(0):
            if not non_exist_tested:
                assert match.need_respond(1)
                non_exist_tested = True
                req = match.requests[0].copy(deep=True)
                assert match.check_request_exist(req)
                resp = SwitchCardResponse(request=req, card_idxs=[-1])  # type: ignore
                with pytest.raises(ValueError):
                    match.respond(resp)
                resp.card_idxs = []
                resp.request.player_idx = 2
                with pytest.raises(ValueError):
                    match.respond(resp)
            resp = agent_0.generate_response(match)
            assert resp is not None
            match.respond(resp)
        elif match.need_respond(1):
            resp = agent_1.generate_response(match)
            assert resp is not None
            match.respond(resp)
        if len(match.requests) == 0:
            match.step(run_continuously=False)


def test_create_dice():
    match = Match()
    act = CreateDiceAction(player_idx=0, number=4, random=True, different=True)
    with pytest.raises(ValueError):
        match._act(act)
    act = CreateDiceAction(
        player_idx=0,
        number=4,
    )
    with pytest.raises(ValueError):
        match._act(act)
    act = CreateDiceAction(
        player_idx=0,
        number=8,
        different=True,
    )
    with pytest.raises(ValueError):
        # different over maximum color number
        match._act(act)
    act = CreateDiceAction(
        player_idx=0,
        number=7,
        different=True,
    )
    match._act(act)
    assert len(match.player_tables[0].dice.colors) == 7
    assert len(set(match.player_tables[0].dice.colors)) == 7
    match.player_tables[0].dice.colors.clear()
    for i in range(50):
        act = CreateDiceAction(
            player_idx=0,
            number=3,
            different=True,
        )
        match._act(act)
        assert len(match.player_tables[0].dice.colors) == 3
        assert len(set(match.player_tables[0].dice.colors)) == 3
        for color in match.player_tables[0].dice.colors:
            assert color != "OMNI"
        match.player_tables[0].dice.colors.clear()


def test_id_wont_duplicate():
    s = set()
    position = ObjectPosition(
        player_idx=-1,
        area=ObjectPositionType.INVALID,
        id=-1,
    )
    for _ in range(50000):
        o = ObjectBase(
            name="Test object",
            position=position,
        )
        assert o.id not in s
        s.add(o.id)


def test_remove_non_exist_equip():
    match = Match()
    deck = Deck.from_str(
        """
        default_version:4.0
        character:GeoMob*3
        Strategize*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    assert match.start()[0]
    with pytest.raises(AssertionError):
        match._action_move_object(
            MoveObjectAction(
                object_position=ObjectPosition(
                    player_idx=0,
                    character_idx=0,
                    area=ObjectPositionType.CHARACTER,
                    id=0,
                ),
                target_position=ObjectPosition(
                    player_idx=0,
                    area=ObjectPositionType.HAND,
                    id=0,
                ),
            )
        )


def test_immutable_position():
    position = ObjectPosition(player_idx=0, area=ObjectPositionType.SUMMON, id=0)
    with pytest.raises(AttributeError):
        position.player_idx = 1
    with pytest.raises(AttributeError):
        del position.area


def test_get_non_exist_object():
    match = Match()
    deck_str = """
    character:Nahida
    Strategize*30
    """
    deck = Deck.from_str(deck_str)
    match.config.character_number = None
    match.config.max_same_card_number = 999
    match.set_deck([deck, deck])
    assert match.start()[0]
    match.step()
    assert (
        match.get_object(
            ObjectPosition(player_idx=0, id=0, area=ObjectPositionType.DECK)
        )
        is None
    )
    assert (
        match.get_object(
            ObjectPosition(player_idx=0, id=0, area=ObjectPositionType.SUPPORT)
        )
        is None
    )


def test_vanilla_element_infusion_team_status():
    class EIC_3_3(ElementalInfusionTeamStatus):
        version: Literal["3.3"] = "3.3"
        position: ObjectPosition = ObjectPosition(
            player_idx=-1,
            area=ObjectPositionType.INVALID,
            id=-1,
        )
        usage: int = 2
        max_usage: int = 2

    desc: Dict[str, DescDictType] = {
        "TEAM_STATUS/Electro Elemental Infusion": {
            "names": {
                "zh-CN": "雷元素附魔",
                "en-US": "Electro Elemental Infusion",
            },
            "descs": {
                "3.3": {
                    "zh-CN": "雷元素附魔",
                    "en-US": "Electro Elemental Infusion",
                },
            },
        },
        "TEAM_STATUS/Geo Elemental Infusion": {
            "names": {
                "zh-CN": "岩元素附魔",
                "en-US": "Geo Elemental Infusion",
            },
            "descs": {
                "3.3": {
                    "zh-CN": "岩元素附魔",
                    "en-US": "Geo Elemental Infusion",
                },
            },
        },
        "TEAM_STATUS/Pyro Elemental Infusion": {
            "names": {
                "zh-CN": "火元素附魔",
                "en-US": "Pyro Elemental Infusion",
            },
            "descs": {
                "3.3": {
                    "zh-CN": "火元素附魔",
                    "en-US": "Pyro Elemental Infusion",
                },
            },
        },
        "TEAM_STATUS/Hydro Elemental Infusion": {
            "names": {
                "zh-CN": "水元素附魔",
                "en-US": "Hydro Elemental Infusion",
            },
            "descs": {
                "3.3": {
                    "zh-CN": "水元素附魔",
                    "en-US": "Hydro Elemental Infusion",
                },
            },
        },
        "TEAM_STATUS/Cryo Elemental Infusion": {
            "names": {
                "zh-CN": "冰元素附魔",
                "en-US": "Cryo Elemental Infusion",
            },
            "descs": {
                "3.3": {
                    "zh-CN": "冰元素附魔",
                    "en-US": "Cryo Elemental Infusion",
                },
            },
        },
        "TEAM_STATUS/Anemo Elemental Infusion": {
            "names": {
                "zh-CN": "风元素附魔",
                "en-US": "Anemo Elemental Infusion",
            },
            "descs": {
                "3.3": {
                    "zh-CN": "风元素附魔",
                    "en-US": "Anemo Elemental Infusion",
                },
            },
        },
        "TEAM_STATUS/Dendro Elemental Infusion": {
            "names": {
                "zh-CN": "草元素附魔",
                "en-US": "Dendro Elemental Infusion",
            },
            "descs": {
                "3.3": {
                    "zh-CN": "草元素附魔",
                    "en-US": "Dendro Elemental Infusion",
                },
            },
        },
    }
    register_class(EIC_3_3, desc)
    electro_ele_status = get_instance(
        TeamStatusBase, {"name": "Electro Elemental Infusion"}
    )
    assert electro_ele_status.infused_elemental_type == DamageElementalType.ELECTRO
    geo_ele_status = get_instance(TeamStatusBase, {"name": "Geo Elemental Infusion"})
    assert geo_ele_status.infused_elemental_type == DamageElementalType.GEO
    pyro_ele_status = get_instance(TeamStatusBase, {"name": "Pyro Elemental Infusion"})
    assert pyro_ele_status.infused_elemental_type == DamageElementalType.PYRO
    hydro_ele_status = get_instance(
        TeamStatusBase, {"name": "Hydro Elemental Infusion"}
    )
    assert hydro_ele_status.infused_elemental_type == DamageElementalType.HYDRO
    cryo_ele_status = get_instance(TeamStatusBase, {"name": "Cryo Elemental Infusion"})
    assert cryo_ele_status.infused_elemental_type == DamageElementalType.CRYO
    anemo_ele_status = get_instance(
        TeamStatusBase, {"name": "Anemo Elemental Infusion"}
    )
    assert anemo_ele_status.infused_elemental_type == DamageElementalType.ANEMO
    dendro_ele_status = get_instance(
        TeamStatusBase, {"name": "Dendro Elemental Infusion"}
    )
    assert dendro_ele_status.infused_elemental_type == DamageElementalType.DENDRO


def test_blacklist_draw_card():
    match = Match()
    deck_str = """
    character:Nahida
    Strategize*29
    Fresh Wind of Freedom
    """
    deck = Deck.from_str(deck_str)
    match.config.character_number = None
    match.config.max_same_card_number = 999
    match.set_deck([deck, deck])
    assert match.start()[0]
    match.step()
    match.respond(
        SwitchCardResponse(
            request=match.requests[0],  # type: ignore
            card_idxs=[1, 2, 3, 4],
        )
    )
    match.respond(
        SwitchCardResponse(
            request=match.requests[0],  # type: ignore
            card_idxs=[1, 2, 3, 4],
        )
    )
    assert len(match.player_tables[0].hands) == 5
    assert len(match.player_tables[1].hands) == 5
    for i in range(1, 5):
        assert match.player_tables[0].hands[i].name == "Strategize"
        assert match.player_tables[1].hands[i].name == "Strategize"
    # then test whitelist draw card but not enough
    match._action_draw_card(
        DrawCardAction(
            player_idx=0,
            number=2,
            whitelist_names=["Fresh Wind of Freedom"],
            draw_if_filtered_not_enough=False,
        )
    )
    assert len(match.player_tables[0].hands) == 5
    match._action_draw_card(
        DrawCardAction(
            player_idx=0,
            number=2,
            whitelist_names=["Fresh Wind of Freedom"],
            draw_if_filtered_not_enough=True,
        )
    )
    assert len(match.player_tables[0].hands) == 7


if __name__ == "__main__":
    # test_object_position_validation()
    # test_match_config_and_match_errors()
    # test_objectbase_wrong_handler_name()
    # test_create_dice()
    # test_id_wont_duplicate()
    # test_remove_non_exist_equip()
    test_vanilla_element_infusion_team_status()
    test_blacklist_draw_card()



from lpsim.server.deck import Deck
from tests.utils_for_test import remove_ids


def test_deck_string():
    deck_dict_1 = Deck(
        **{
            "name": "Deck",
            "characters": [
                {
                    "name": "DendroMobMage",
                    "element": "DENDRO",
                },
                {
                    "name": "DendroMobMage",
                    "element": "DENDRO",
                },
                {
                    "name": "ElectroMobMage",
                    "element": "ELECTRO",
                },
            ],
            "cards": [
                {
                    "name": "Strategize",
                }
            ]
            * 30,
            "default_version": "4.0",
        }
    )
    deck_str_1 = Deck.from_str(
        """
        default_version:4.0
        character:DendroMobMage*2
        character:ElectroMobMage
        Strategize*30

    """
    )
    deck_dict_1 = remove_ids(deck_dict_1)
    deck_str_1 = remove_ids(deck_str_1)
    assert deck_str_1 == deck_dict_1
    deck_dict_2 = Deck(
        **{
            "name": "Deck",
            "characters": [
                {
                    "name": "DendroMobMage",
                    "element": "DENDRO",
                },
                {
                    "name": "DendroMobMage",
                    "element": "DENDRO",
                },
                {
                    "name": "ElectroMobMage",
                    "element": "ELECTRO",
                },
            ],
            "cards": [
                {
                    "name": "Wine-Stained Tricorne",
                }
            ]
            * 10
            + [
                {
                    "name": "Laurel Coronet",
                }
            ]
            * 10
            + [
                {
                    "name": "Strategize",
                }
            ]
            * 10,
            "default_version": "4.0",
        }
    )
    deck_str_2 = Deck.from_str(
        """
        default_version:4.0

        # characters
        character:DendroMobMage*2
        character:ElectroMobMage

        # cards
        Wine-Stained Tricorne*10
        Laurel Coronet*10
        Strategize*10
    """
    )
    deck_dict_2 = remove_ids(deck_dict_2)
    deck_str_2 = remove_ids(deck_str_2)
    assert deck_str_2 == deck_dict_2
    deck_dict_3 = Deck(
        **{
            "name": "Deck",
            "characters": [
                {
                    "name": "Fischl",
                },
                {
                    "name": "DendroMobMage",
                    "element": "DENDRO",
                },
                {
                    "name": "PyroMobMage",
                    "element": "PYRO",
                },
            ],
            "cards": [
                {
                    "name": "Stellar Predator",
                }
            ]
            * 30,
            "default_version": "4.0",
        }
    )
    deck_str_3 = Deck.from_str(
        """
        default_version:4.0
        character:Fischl
        character:DendroMobMage
        character:PyroMobMage
        Stellar Predator*30
    """
    )
    deck_dict_3 = remove_ids(deck_dict_3)
    deck_str_3 = remove_ids(deck_str_3)
    assert deck_str_3 == deck_dict_3


def test_deck_assertions():
    deck = Deck.from_str(
        """
        default_version:4.0
        character:DendroMobMage*2
        character:ElectroMobMage
        Strategize*30
    """
    )
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=False,
    )[0]
    assert deck.check_legal(
        card_number=30,
        max_same_card_number=30,
        character_number=3,
        check_restriction=False,
    )[0]
    assert not deck.check_legal(
        card_number=29,
        max_same_card_number=30,
        character_number=3,
        check_restriction=False,
    )[0]
    assert not deck.check_legal(
        card_number=31,
        max_same_card_number=30,
        character_number=3,
        check_restriction=False,
    )[0]
    assert not deck.check_legal(
        card_number=30,
        max_same_card_number=29,
        character_number=3,
        check_restriction=False,
    )[0]
    assert deck.check_legal(
        card_number=30,
        max_same_card_number=31,
        character_number=3,
        check_restriction=False,
    )[0]
    assert not deck.check_legal(
        card_number=30,
        max_same_card_number=31,
        character_number=2,
        check_restriction=False,
    )[0]
    assert not deck.check_legal(
        card_number=30,
        max_same_card_number=31,
        character_number=4,
        check_restriction=False,
    )[0]


def test_deck_restriction():
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Collei
        character:Barbara
        Wind and Freedom*30
    """
    )
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=False,
    )[0]
    assert not deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=True,
    )[0]
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Mona
        character:Barbara
        Wind and Freedom*30
        Sweet Madame*30
    """
    )
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=False,
    )[0]
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=True,
    )[0]
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Mona
        character:Barbara
        Elemental Resonance: Woven Waters*30
    """
    )
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=False,
    )[0]
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=True,
    )[0]
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Mona
        character:Barbara
        Elemental Resonance: Woven Weeds*30
    """
    )
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=False,
    )[0]
    assert not deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=True,
    )[0]
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Mona
        character:Barbara
        The Seed of Stored Knowledge*5
        Prophecy of Submersion*5
        Glorious Season*5
    """
    )
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=False,
    )[0]
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=True,
    )[0]
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Mona
        character:Barbara
        The Seed of Stored Knowledge*5
        Prophecy of Submersion*5
        Grand Expectation*5
    """
    )
    assert deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=False,
    )[0]
    assert not deck.check_legal(
        card_number=None,
        max_same_card_number=None,
        character_number=None,
        check_restriction=True,
    )[0]


def test_deck_to_str_to_deck():
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida
        character:Mona
        character:Barbara
        The Seed of Stored Knowledge*5
        Prophecy of Submersion*5
        Grand Expectation*5
    """
    )
    deck2 = Deck.from_str(deck.to_str())
    assert remove_ids(deck) == remove_ids(deck2)


if __name__ == "__main__":
    test_deck_string()
    test_deck_restriction()
    test_deck_to_str_to_deck()



from typing import Dict, List, Literal

from .....utils.class_registry import register_class

from ....action import Actions, ChangeObjectUsageAction, DrawCardAction

from ....status.team_status.base import ExtraAttackTeamStatus, RoundTeamStatus
from ....modifiable_values import DamageElementEnhanceValue, InitialDiceColorValue
from ....event import RoundEndEventArguments, SkillEndEventArguments
from ....status.character_status.base import ElementalInfusionCharacterStatus
from ....match import Match
from ....struct import Cost
from ....consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    SkillType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class BreakthroughStatus_4_3(ElementalInfusionCharacterStatus):
    """
    Adding usage ways:
    1. when create, it contains 1.
    2. when renew, it adds 1 (by ADD renew_type)
    3. added by elemental skill (by skill add usage action)
    4. added in round end
    """

    name: Literal["Breakthrough"] = "Breakthrough"
    version: Literal["4.3"] = "4.3"
    usage: int = 1
    max_usage: int = 3
    infused_elemental_type: ElementType = ElementType.HYDRO
    icon_type: Literal[IconType.ELEMENT_ENCHANT_WATER] = IconType.ELEMENT_ENCHANT_WATER

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[Actions]:
        self.usage = min(self.usage + 1, self.max_usage)
        return []

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Match,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        if self.usage < 2:
            # usage not enough, not modify
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(value, match, mode)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[DrawCardAction]:
        """
        If self use normal attack and have usage, (it should be enhanced),
        draw a card.
        """
        if self.usage < 2:
            # usage not enough, not modify
            return []
        if (
            event.action.skill_type != SkillType.NORMAL_ATTACK
            or event.action.position.player_idx != self.position.player_idx
            or event.action.position.character_idx != self.position.character_idx
        ):
            # not self use normal attack
            return []
        # decrease usage and draw a card
        self.usage -= 2
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ]


class ExquisiteThrow_4_3(ExtraAttackTeamStatus, RoundTeamStatus):
    name: Literal["Exquisite Throw"] = "Exquisite Throw"
    version: Literal["4.3"] = "4.3"
    damage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    usage: int = 2
    max_usage: int = 2
    decrease_usage: bool = False
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS
    trigger_skill_type: SkillType = SkillType.NORMAL_ATTACK


class LingeringLifeline(ElementalSkillBase):
    name: Literal["Lingering Lifeline"] = "Lingering Lifeline"
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if s.name == "Breakthrough":  # pragma: no branch
                break_through = s
                break
        else:
            raise AssertionError("Breakthrough not found")
        return super().get_actions(match) + [
            ChangeObjectUsageAction(
                object_position=break_through.position,
                change_usage=2,
                max_usage=break_through.max_usage,
            )
        ]


class DepthClarionDice(ElementalBurstBase):
    name: Literal["Depth-Clarion Dice"] = "Depth-Clarion Dice"
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
        charge=3,
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status("Exquisite Throw"),
        ]


class Breakthrough(CreateStatusPassiveSkill):
    name: Literal["Breakthrough"] = "Breakthrough"
    status_name: Literal["Breakthrough"] = "Breakthrough"


class TurnControl_4_3(SkillTalent):
    name: Literal["Turn Control"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Yelan"] = "Yelan"
    skill: Literal["Lingering Lifeline"] = "Lingering Lifeline"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
    )

    def value_modifier_INITIAL_DICE_COLOR(
        self, value: InitialDiceColorValue, match: Match, mode: Literal["REAL", "TEST"]
    ) -> InitialDiceColorValue:
        """
        If equipped, add omni based on element type number.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped, not modify
            return value
        if value.player_idx != self.position.player_idx:
            # not self player, not modify
            return value
        characters = match.player_tables[self.position.player_idx].characters
        elements = set()
        for character in characters:
            elements.add(character.element)
        # add omni based on element type number
        value.dice_colors += [DieColor.OMNI] * len(elements)
        return value


class Yelan_4_3(CharacterBase):
    name: Literal["Yelan"]
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | LingeringLifeline | DepthClarionDice | Breakthrough
    ] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Stealthy Bowshot",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            LingeringLifeline(),
            DepthClarionDice(),
            Breakthrough(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Yelan": {
        "names": {"en-US": "Yelan", "zh-CN": "夜兰"},
        "image_path": "cardface/Char_Avatar_Yelan.png",  # noqa: E501
        "id": 1209,
        "descs": {
            "4.3": {"en-US": "", "zh-CN": ""},
        },
    },
    "SKILL_Yelan_NORMAL_ATTACK/Stealthy Bowshot": {
        "names": {"en-US": "Stealthy Bowshot", "zh-CN": "潜形隐曜弓"},
        "descs": {
            "4.3": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Yelan_ELEMENTAL_SKILL/Lingering Lifeline": {
        "names": {"en-US": "Lingering Lifeline", "zh-CN": "萦络纵命索"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Hydro DMG. This character gains 2 Breakthrough stacks.",  # noqa: E501
                "zh-CN": "造成3点水元素伤害，此角色的破局层数+2。",
            }
        },
    },
    "CHARACTER_STATUS/Breakthrough": {
        "names": {"en-US": "Breakthrough", "zh-CN": "破局"},
        "descs": {
            "4.3": {
                "en-US": "This state starts with 1 Breakthrough stack. When re-attached, add 1 Breakthrough stack. Breakthrough can have a maximum of 3 stacks.\nEnd Phase: Gain 1 Breakthrough stack.\n\nWhen the character to which this is attached performs a $[K51]: If possible, consume 2 Breakthrough stacks to convert Physical DMG dealt into Hydro DMG and draw 1 card.",  # noqa: E501
                "zh-CN": "此状态初始具有1层「破局」；重复附属时，叠加1层「破局」。「破局」最多可以叠加到3层。\n结束阶段：叠加1层「破局」。\n\n所附属角色普通攻击时：如可能，则消耗2层「破局」，使造成的物理伤害转换为水元素伤害，并抓1张牌。",  # noqa: E501
            }
        },
    },
    "SKILL_Yelan_ELEMENTAL_BURST/Depth-Clarion Dice": {
        "names": {"en-US": "Depth-Clarion Dice", "zh-CN": "渊图玲珑骰"},
        "descs": {
            "4.3": {
                "en-US": "Deals 1 Hydro DMG, creates 1 Exquisite Throw.",
                "zh-CN": "造成1点水元素伤害，生成玄掷玲珑。",
            }
        },
    },
    "TEAM_STATUS/Exquisite Throw": {
        "names": {"en-US": "Exquisite Throw", "zh-CN": "玄掷玲珑"},
        "descs": {
            "4.3": {
                "en-US": "After your character uses a Normal Attack: Deal 1 Hydro DMG. If the remaining Duration (Rounds) is only 1, this DMG increases by 1.\nDuration (Rounds): 2",  # noqa: E501
                "zh-CN": "我方角色普通攻击后：造成2点水元素伤害。\n持续回合：2",
            }
        },
        "image_path": "status/Yelan_E.png",
    },
    "SKILL_Yelan_PASSIVE/Breakthrough": {
        "names": {"en-US": "Breakthrough", "zh-CN": "破局"},
        "descs": {
            "4.3": {
                "en-US": "(Passive) When the battle begins, this character gains Breakthrough.",  # noqa: E501
                "zh-CN": "【被动】战斗开始时，初始附属破局。",
            }
        },
    },
    "TALENT_Yelan/Turn Control": {
        "names": {"en-US": "Turn Control", "zh-CN": "猜先有方"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Yelan, equip this card.\nAfter Yelan equips this card, immediately use Lingering Lifeline once.\nRoll Phase: If Yelan, who has this card equipped, is on the field, 1 Elemental Dice will be rolled as Omni Element for each Elemental Type in your party. (Max 3)\n(You must have Yelan in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为夜兰时，装备此牌。\n夜兰装备此牌后，立刻使用一次萦络纵命索。\n投掷阶段：装备有此牌的夜兰在场，则我方队伍中每有1种元素类型，就使1个元素骰总是投出万能元素。（最多3个）\n（牌组中包含夜兰，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Yelan.png",  # noqa: E501
        "id": 212091,
    },
}


register_class(
    BreakthroughStatus_4_3 | ExquisiteThrow_4_3 | TurnControl_4_3 | Yelan_4_3, desc
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class


from ...modifiable_values import DamageValue
from ...event import (
    CharacterReviveEventArguments,
    CreateObjectEventArguments,
    GameStartEventArguments,
    RoundEndEventArguments,
    SkillEndEventArguments,
)
from ...action import Actions, CreateObjectAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PassiveSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillBase,
    SkillTalent,
)


# Skills


class AttachRiptideWhenSkillEnd(SkillBase):
    version: str

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When skill end, if is this skill, and target
        is alive, attach riptide to target.
        """
        if event.action.position.id != self.id:
            # not self skill
            return []
        target = match.player_tables[1 - self.position.player_idx].characters[
            event.action.target_position.character_idx
        ]
        if target.is_defeated:
            # target is defeated
            return []
        # attach riptide to target
        return [
            CreateObjectAction(
                object_name="Riptide",
                object_position=event.action.target_position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={"version": self.version},
            )
        ]


class FoulLegacyRagingTide(ElementalSkillBase, AttachRiptideWhenSkillEnd):
    name: Literal["Foul Legacy: Raging Tide"] = "Foul Legacy: Raging Tide"
    version: Literal["4.1"] = "4.1"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object then attack
        """
        return [
            # Tartaglia need to create the status first, otherwise cannot
            # receive the damage increase
            self.create_character_status("Melee Stance", {"version": self.version}),
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.charge_self(1),
        ]


class HavocObliteration(ElementalBurstBase, AttachRiptideWhenSkillEnd):
    name: Literal["Havoc: Obliteration"] = "Havoc: Obliteration"
    version: Literal["4.1"] = "4.1"
    damage: int = 7
    ranged_damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
        charge=3,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        ranged_found: bool = False
        melee_found: bool = False
        for status in character.status:
            if status.name == "Ranged Stance":
                ranged_found = True
            elif status.name == "Melee Stance":  # pragma: no branch
                melee_found = True
        if ranged_found and melee_found:
            raise AssertionError("Cannot be both ranged and melee stance")
        if not (ranged_found or melee_found):
            raise AssertionError("Must be ranged or melee stance")
        if ranged_found:
            # ranged damage, and reclaim 2 energy. riptide will attach in
            # skill end.
            return [
                self.charge_self(-3),
                self.attack_opposite_active(
                    match, self.ranged_damage, DamageElementalType.HYDRO
                ),
                self.charge_self(2),
            ]
        else:
            # 7 damage
            return super().get_actions(match)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If ranged stance, call super() to attach riptide.
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        ranged_found: bool = False
        for status in character.status:
            if status.name == "Ranged Stance":
                ranged_found = True
                break
        if not ranged_found:
            # not ranged stance
            return []
        return super().event_handler_SKILL_END(event, match)


class TideWithholder(PassiveSkillBase):
    """
    Only generate Ranged Stance at game start, and generate riptide event
    handler for system. status switch will be controlled by two status
    themself.
    """

    name: Literal["Tide Withholder"] = "Tide Withholder"
    status_name: Literal["Ranged Stance"] = "Ranged Stance"
    version: Literal["4.1"] = "4.1"

    def event_handler_CHARACTER_REVIVE(
        self, event: CharacterReviveEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When character revive, gain status
        """
        if (
            event.action.player_idx != self.position.player_idx
            or event.action.character_idx != self.position.character_idx
        ):
            # not this character, do nothing
            return []
        return [
            self.create_character_status("Ranged Stance", {"version": self.version}),
        ]

    def event_handler_GAME_START(
        self, event: GameStartEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When game begin, gain ranged stance and system handler
        """
        return [
            self.create_character_status("Ranged Stance", {"version": self.version}),
        ]


# Talents


class AbyssalMayhemHydrospout_4_1(SkillTalent):
    name: Literal["Abyssal Mayhem: Hydrospout"]
    version: Literal["4.1"] = "4.1"
    character_name: Literal["Tartaglia"] = "Tartaglia"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO,
        elemental_dice_number=3,
    )
    skill: Literal["Foul Legacy: Raging Tide"] = "Foul Legacy: Raging Tide"
    only_active_character: bool = True

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        Make 1 damage to character if affected by Riptide. When only active
        character (after 4.1), only check active character. Otherwise,
        check all characters.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        ret: List[MakeDamageAction] = []
        active_idx = match.player_tables[
            1 - self.position.player_idx
        ].active_character_idx
        for cidx, character in enumerate(
            match.player_tables[1 - self.position.player_idx].characters
        ):
            if self.only_active_character and cidx != active_idx:
                continue
            for status in character.status:
                if status.name == "Riptide":
                    ret.append(
                        MakeDamageAction(
                            damage_value_list=[
                                DamageValue(
                                    position=self.position,
                                    damage_type=DamageType.DAMAGE,
                                    target_position=character.position,
                                    damage=1,
                                    damage_elemental_type=DamageElementalType.PIERCING,
                                    cost=Cost(),
                                )
                            ]
                        )
                    )
                    break
        return ret


# character base


class Tartaglia_4_1(CharacterBase):
    name: Literal["Tartaglia"]
    desc: Literal["", "transform"] = ""
    version: Literal["4.1"] = "4.1"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase
        | FoulLegacyRagingTide
        | HavocObliteration
        | TideWithholder
    ] = []
    faction: List[FactionType] = [FactionType.FATUI]
    weapon_type: WeaponType = WeaponType.BOW

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Cutting Torrent",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            FoulLegacyRagingTide(),
            HavocObliteration(),
            TideWithholder(),
        ]

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        status = (
            match.player_tables[self.position.player_idx]
            .characters[self.position.character_idx]
            .status
        )
        for s in status:
            if s.name == "Melee Stance":
                self.desc = "transform"
                return []
        self.desc = ""
        return []


register_class(Tartaglia_4_1 | AbyssalMayhemHydrospout_4_1)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...modifiable_values import DamageIncreaseValue
from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class GuideToAfterlife(ElementalSkillBase):
    name: Literal["Guide to Afterlife"] = "Guide to Afterlife"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=2)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        gain charge and create status
        """
        return [
            self.charge_self(1),
            self.create_character_status("Paramita Papilio"),
        ]


class SpiritSoother(ElementalBurstBase):
    name: Literal["Spirit Soother"] = "Spirit Soother"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.PYRO, elemental_dice_number=3, charge=3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Check if hp <= 6. If so, increase damage and heal
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        heal = 2
        if character.hp <= 6:
            self.damage = 5
            heal = 3
        damage_action = self.attack_opposite_active(
            match, self.damage, self.damage_type
        )
        heal_action = self.attack_self(match, -heal)
        damage_action.damage_value_list += heal_action.damage_value_list
        self.damage = 4
        return [
            self.charge_self(-3),
            damage_action,
        ]


# Talents


class SanguineRouge_3_7(SkillTalent):
    name: Literal["Sanguine Rouge"]
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Hu Tao"] = "Hu Tao"
    cost: Cost = Cost(elemental_dice_color=DieColor.PYRO, elemental_dice_number=2)
    skill: Literal["Guide to Afterlife"] = "Guide to Afterlife"

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not self use damage skill
            return value
        if value.damage_elemental_type != DamageElementalType.PYRO:
            # not pyro damage
            return value
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp > 6:
            # hp > 6
            return value
        # increase damage
        value.damage += 1
        return value


# character base


class HuTao_3_7(CharacterBase):
    name: Literal["Hu Tao"]
    version: Literal["3.7"] = "3.7"
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[PhysicalNormalAttackBase | GuideToAfterlife | SpiritSoother] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Secret Spear of Wangsheng",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            GuideToAfterlife(),
            SpiritSoother(),
        ]


register_class(HuTao_3_7 | SanguineRouge_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...struct import Cost, ObjectPosition

from ...modifiable_values import CostValue, DamageElementEnhanceValue, DamageValue

from ...consts import (
    CostLabels,
    DamageElementalType,
    DamageType,
    ElementType,
    ElementalReactionType,
    IconType,
    ObjectPositionType,
    SkillType,
)

from ...action import (
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    MakeDamageAction,
    RemoveObjectAction,
)

from ...event import (
    AfterMakeDamageEventArguments,
    ReceiveDamageEventArguments,
    RoundEndEventArguments,
    SkillEndEventArguments,
)
from .base import (
    CharacterStatusBase,
    ElementalInfusionCharacterStatus,
    UsageCharacterStatus,
)


class SeedOfSkandha_3_7(UsageCharacterStatus):
    name: Literal["Seed of Skandha"] = "Seed of Skandha"
    version: Literal["3.7"] = "3.7"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        Only seed of character received the damage will trigger event, it will
        check all character status and trigger them and call ChangeObjectUsage.
        """
        damage_value = event.final_damage
        if damage_value.damage_type != DamageType.DAMAGE:
            # not damage, not trigger
            return []
        if damage_value.element_reaction == ElementalReactionType.NONE:
            # not elemental reaction, not trigger
            return []
        if not self.position.check_position_valid(
            damage_value.target_position,
            match,
            player_idx_same=True,
            character_idx_same=True,
        ):
            # damage not received by self, not trigger
            return []
        assert self.usage > 0
        # trigger, check all seed on the same side
        damage_actions: List[MakeDamageAction] = []
        change_usage_actions: List[ChangeObjectUsageAction] = []
        table = match.player_tables[self.position.player_idx]
        has_pyro_character = False
        has_talent = False
        # check if enemy has pyro character and has talent nahida
        for character in match.player_tables[1 - self.position.player_idx].characters:
            if character.element == ElementType.PYRO:
                has_pyro_character = True
            if character.name == "Nahida":
                if character.talent is not None:
                    has_talent = True
        aidx = table.active_character_idx
        characters = table.characters[aidx:] + table.characters[:aidx]
        for character in characters:
            for status in character.status:
                if status.name == "Seed of Skandha":
                    # found a seed, trigger it
                    assert status.usage > 0
                    d_ele_type = DamageElementalType.PIERCING
                    if has_pyro_character and has_talent:
                        # enemy have talent nahida and pyro, dendeo damage
                        d_ele_type = DamageElementalType.DENDRO
                    # change usage first, so no need to claim new trigger
                    change_usage_actions.append(
                        ChangeObjectUsageAction(
                            object_position=status.position,
                            change_usage=-1,
                        )
                    )
                    damage_actions.append(
                        MakeDamageAction(
                            damage_value_list=[
                                DamageValue(
                                    position=status.position,
                                    damage=1,
                                    target_position=character.position,
                                    damage_type=DamageType.DAMAGE,
                                    cost=Cost(),
                                    damage_elemental_type=d_ele_type,
                                )
                            ],
                        )
                    )
        # firstly change usage, to avoid object is disappeared.
        return change_usage_actions + damage_actions


class VijnanaSuffusion_3_6(ElementalInfusionCharacterStatus, UsageCharacterStatus):
    name: Literal["Vijnana Suffusion"] = "Vijnana Suffusion"
    version: Literal["3.6"] = "3.6"
    usage: int = 2
    max_usage: int = 2
    infused_elemental_type: DamageElementalType = DamageElementalType.DENDRO

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        """
        When character use charged attack, change damage type to dendro
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not use normal attack
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, not modify
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(value, match, mode)

    def value_modifier_COST(
        self, value: CostValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> CostValue:
        """
        If talent equipped, and is charged attack, decrease 1 any cost
        """
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.talent is None:
            # no talent, not modify
            return value
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not this character use skill, not modify
            return value
        if not (
            match.player_tables[self.position.player_idx].charge_satisfied
            and value.cost.label & CostLabels.NORMAL_ATTACK.value != 0
        ):
            # not charged attack, not modify
            return value
        # decrease
        value.cost.decrease_cost(None)
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[CreateObjectAction | RemoveObjectAction]:
        """
        If self use charged attack, summon Clusterbloom Arrow
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not same player or character, or not skill, not modify
            return []
        if not (
            match.player_tables[self.position.player_idx].charge_satisfied
            and match.get_object(event.action.position).skill_type
            == SkillType.NORMAL_ATTACK
        ):
            # not charged attack, not modify
            return []
        assert self.usage > 0
        # summon
        self.usage -= 1
        return [
            CreateObjectAction(
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.SUMMON,
                    id=0,
                ),
                object_name="Clusterbloom Arrow",
                object_arguments={},
            )
        ] + self.check_should_remove()


class RadicalVitality_3_3(CharacterStatusBase):
    name: Literal["Radical Vitality"] = "Radical Vitality"
    desc: Literal["", "talent"] = ""
    version: Literal["3.3"] = "3.3"
    usage: int = 0
    max_usage: int = 3
    element_damage_triggered: bool = False
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._update_desc()

    def _update_desc(self) -> None:
        if self.max_usage == 4:
            # talent activated
            self.desc = "talent"
        else:
            self.desc = ""

    def renew(self, new_status: "RadicalVitality_3_3") -> None:
        super().renew(new_status)
        self._update_desc()

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        if self.usage == self.max_usage:
            self.usage = 0
        return []

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        If contains elemental damage and target or source self, gain progress
        """
        satisfied: bool = False
        for damage in event.damages:
            source_self = self.position.check_position_valid(
                damage.final_damage.position,
                match,
                player_idx_same=True,
                character_idx_same=True,
            )
            target_self = self.position.check_position_valid(
                damage.final_damage.target_position,
                match,
                player_idx_same=True,
                character_idx_same=True,
            )
            if source_self or target_self:
                # source or target self
                if damage.final_damage.damage_type != DamageType.DAMAGE:
                    # heal, do nothing
                    continue
                if damage.final_damage.damage_elemental_type in [
                    DamageElementalType.PHYSICAL,
                    DamageElementalType.PIERCING,
                    DamageElementalType.HEAL,
                ]:
                    # physical or piercing or heal, do nothing
                    continue
                if source_self and (
                    damage.final_damage.cost.label & CostLabels.ELEMENTAL_BURST.value
                    != 0
                ):
                    # source self, and elemental burst, do nothing
                    continue
                satisfied = True
        if satisfied:
            # increase usage
            self.usage = min(self.usage + 1, self.max_usage)
        return []


register_class(SeedOfSkandha_3_7 | VijnanaSuffusion_3_6 | RadicalVitality_3_3)



from typing import List, Literal, Any
from pydantic import validator

from ...utils import accept_same_or_higher_version
from ...utils.class_registry import register_base_class
from ..object_base import ObjectBase
from ..consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    ElementType,
    ElementalReactionType,
    IconType,
    ObjectPositionType,
    ObjectType,
    DamageElementalType,
    DamageType,
)
from ..event import (
    CreateObjectEventArguments,
    DeclareRoundEndEventArguments,
    MakeDamageEventArguments,
    ReceiveDamageEventArguments,
    RoundEndEventArguments,
    ChangeObjectUsageEventArguments,
    RoundPrepareEventArguments,
)
from ..action import (
    ActionTypes,
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    MakeDamageAction,
    RemoveObjectAction,
)
from ..modifiable_values import DamageDecreaseValue, DamageValue
from ..struct import Cost, ObjectPosition


class SummonBase(ObjectBase):
    type: Literal[ObjectType.SUMMON] = ObjectType.SUMMON
    renew_type: Literal["ADD", "RESET", "RESET_WITH_MAX"] = "RESET_WITH_MAX"
    name: str
    strict_version_validation: bool = False  # default accept higher versions
    version: str
    usage: int
    max_usage: int
    damage_elemental_type: DamageElementalType
    damage: int

    # icon type is used to show the icon on the summon top right.
    icon_type: Literal[
        IconType.SHIELD, IconType.BARRIER, IconType.TIMESTATE, IconType.COUNTER
    ]
    # when status icon type is not none, it will show in team status area
    status_icon_type: Literal[IconType.NONE] = IconType.NONE

    @validator("version", pre=True)
    def accept_same_or_higher_version(cls, v: str, values):
        return accept_same_or_higher_version(cls, v, values)

    def renew(self, new_status: "SummonBase") -> None:
        """
        Renew the status.
        """
        if self.renew_type == "ADD":
            self.usage += new_status.usage
            if self.max_usage < self.usage:
                self.usage = self.max_usage
        elif self.renew_type == "RESET":
            raise NotImplementedError("RESET is not supported")
            self.usage = new_status.usage
        else:
            assert self.renew_type == "RESET_WITH_MAX"
            self.usage = max(self.usage, new_status.usage)

    def is_valid(self, match: Any) -> bool:
        """
        For summons, it is expected to never be used as card.
        """
        raise AssertionError("SummonBase is not expected to be used as card")


register_base_class(SummonBase)


class AttackerSummonBase(SummonBase):
    """
    Attacker summons, e.g. Guoba, Oz. They do attack on round end, and
    disappears when run out of usage. Specially, Melody Loop can also be a
    Attacker Summon, as it also makes damage with type HEAL.
    """

    name: str
    version: str
    usage: int
    max_usage: int
    damage_elemental_type: DamageElementalType
    damage: int

    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        When round end, make damage to the opponent, and decrease self usage.
        """
        player_idx = self.position.player_idx
        assert self.usage > 0
        damage_type = DamageType.DAMAGE
        target_table = match.player_tables[1 - player_idx]
        if self.damage_elemental_type == DamageElementalType.HEAL:
            damage_type = DamageType.HEAL
            target_table = match.player_tables[player_idx]
        target_character = target_table.get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=damage_type,
                        target_position=target_character.position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_elemental_type,
                        cost=Cost(),
                    )
                ],
            ),
            ChangeObjectUsageAction(object_position=self.position, change_usage=-1),
        ]

    def _remove(self, match: Any) -> List[RemoveObjectAction]:
        """
        Remove the summon.
        """
        return [
            RemoveObjectAction(
                object_position=self.position,
            )
        ]

    def event_handler_CHANGE_OBJECT_USAGE(
        self, event: ChangeObjectUsageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When usage is 0, remove the summon.
        """
        if self.usage <= 0:
            return self._remove(match)
        return []

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[RemoveObjectAction]:
        """
        When usage is 0, remove the summon.
        """
        if self.usage <= 0:
            return self._remove(match)
        return []


class AOESummonBase(AttackerSummonBase):
    """
    Base class that deals AOE damage. It will attack active character
    damage+element, back character back_damage_piercing.
    """

    back_damage: int

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        damage_action = ret[0]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        target_table = match.player_tables[1 - self.position.player_idx]
        characters = target_table.characters
        for cid, character in enumerate(characters):
            if cid == target_table.active_character_idx:
                continue
            if character.is_defeated:
                continue
            damage_action.damage_value_list.append(
                DamageValue(
                    position=self.position,
                    damage_type=DamageType.DAMAGE,
                    target_position=character.position,
                    damage=self.back_damage,
                    damage_elemental_type=DamageElementalType.PIERCING,
                    cost=Cost(),
                )
            )
        return ret


class DeclareRoundEndAttackSummonBase(AttackerSummonBase):
    """
    Summons that will attack when declare round end, e.g. Sesshou Sakura.
    """

    extra_attack_usage: int

    def event_handler_DECLARE_ROUND_END(
        self, event: DeclareRoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        if usage larger than expected, return event_handler_ROUND_END
        """
        if event.action.player_idx != self.position.player_idx:
            # not our player, do nothing
            return []
        if self.usage < self.extra_attack_usage:
            # not enough usage, do nothing
            return []
        # as attack not use event, use fake event instead.
        fake_event: Any = None
        # make damage
        return self.event_handler_ROUND_END(fake_event, match)


class DefendSummonBase(SummonBase):
    """
    Defend summons, e.g. Ushi, Frog, Baron Bunny. decrease damage with its rule
    when receive damage, and decrease usage by 1, and will do one-time damage
    in round end.

    Args:
        name: The name of the summon.
        usage: The usage of the summon, which is shown on top right when it is
            summoned, represents the number of times it can decrease damage.
        max_usage: The maximum usage of the summon.
        damage_elemental_type: The elemental type when attack.
        damage: The damage when attack.
        attack_until_run_out_of_usage: Whether attack until run out of usage.
            If true, when usage is not zero, it will remain on the summon area
            and do nothing in round end. If false, when usage is not zero, it
            will do damage in round end and disappear.
        min_damage_to_trigger: The minimum damage to trigger the damage
            decrease.
        max_in_one_time: the maximum damage decrease in one time.
        decrease_usage_by_damage: Always be false for defend summons.
    """

    name: str
    version: str
    usage: int
    max_usage: int

    # attack params
    damage_elemental_type: DamageElementalType
    damage: int

    # defend params
    attack_until_run_out_of_usage: bool
    min_damage_to_trigger: int
    max_in_one_time: int
    decrease_usage_by_damage: bool = False

    icon_type: Literal[IconType.BARRIER] = IconType.BARRIER

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | RemoveObjectAction]:
        """
        When round end, make damage to the opponent if is needed and remove
        itself.
        """
        player_idx = self.position.player_idx
        if self.usage > 0 and self.attack_until_run_out_of_usage:
            # attack until run out of usage
            return []
        target_table = match.player_tables[1 - player_idx]
        target_character = target_table.get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target_character.position,
                        damage=self.damage,
                        damage_elemental_type=self.damage_elemental_type,
                        cost=Cost(),
                    )
                ],
            )
        ] + self._remove(match)

    def _remove(self, match: Any) -> List[RemoveObjectAction]:
        """
        Remove the summon.
        """
        return [
            RemoveObjectAction(
                object_position=self.position,
            )
        ]

    def value_modifier_DAMAGE_DECREASE(
        self, value: DamageDecreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageDecreaseValue:
        """
        Decrease damage with its rule, and decrease usage.
        """
        if not value.is_corresponding_character_receive_damage(
            self.position,
            match,
        ):
            # not this character receive damage, not modify
            return value
        new_usage = value.apply_shield(
            self.usage,
            self.min_damage_to_trigger,
            self.max_in_one_time,
            self.decrease_usage_by_damage,
        )
        assert mode == "REAL"
        self.usage = new_usage
        return value


class ShieldSummonBase(DefendSummonBase):
    """
    Shield summons, which decrease damage like yellow shield, decrease damage
    by its usage and decrease corresponding usage. Currently no such summon.

    Args:
        name: The name of the summon.
        usage: The usage of the summon, which is shown on top right when it is
            summoned, represents the number of times it can decrease damage.
        max_usage: The maximum usage of the summon.
        damage_elemental_type: The elemental type when attack.
        damage: The damage when attack.
        attack_until_run_out_of_usage: Whether attack until run out of usage.
            If true, when usage is not zero, it will remain on the summon area
            and do nothing in round end. If false, when usage is not zero, it
            will do damage in round end and disappear.

        decrease_usage_by_damage: Always be true for shield summons.
        min_damage_to_trigger: Not used, always be 0 for shield summons.
        max_in_one_time: Not used, always be 0 for shield summons.
    """

    name: str
    version: str
    usage: int
    max_usage: int
    damage_elemental_type: DamageElementalType
    damage: int
    attack_until_run_out_of_usage: bool

    icon_type: Literal[IconType.SHIELD] = IconType.SHIELD

    # papams passing to apply_shield is fixed
    min_damage_to_trigger: int = 0
    max_in_one_time: int = 0
    decrease_usage_by_damage: bool = True


class SwirlChangeSummonBase(AttackerSummonBase):
    """
    Base class for summons that can change damage elemental type when swirl
    reaction made by our character or summon.
    Note: create summon before attack, so it can change element immediately.
    """

    name: str
    version: str
    usage: int
    max_usage: int

    damage_elemental_type: DamageElementalType = DamageElementalType.ANEMO

    def renew(self, new_status: SummonBase) -> None:
        """
        When renew, also renew damage elemental type.
        """
        super().renew(new_status)
        self.damage_elemental_type = DamageElementalType.ANEMO

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        When anyone receives damage, and has swirl reaction, and made by
        our character or summon, and current damage type is anemo,
        change stormeye damage element.
        """
        if self.damage_elemental_type != DamageElementalType.ANEMO:
            # already changed, do nothing
            return []
        if event.final_damage.position.area not in [
            ObjectPositionType.SKILL,
            ObjectPositionType.SUMMON,
        ]:  # pragma: no cover
            # not character or summon made damage, do nothing
            return []
        if event.final_damage.position.player_idx != self.position.player_idx:
            # not our player made damage, do nothing
            return []
        if event.final_damage.element_reaction != ElementalReactionType.SWIRL:
            # not swirl reaction, do nothing
            return []
        elements = event.final_damage.reacted_elements
        assert elements[0] == ElementType.ANEMO, "First element must be anemo"
        # do type change
        self.damage_elemental_type = ELEMENT_TO_DAMAGE_TYPE[elements[1]]
        return []


class AttackAndGenerateStatusSummonBase(AttackerSummonBase):
    """
    Summons that will attack opposite, and generate status to self periodly,
    usually defend status. e.g. Dehya, Lynette. When the summon is removed,
    it will also try to remove the status.
    """

    name: str
    version: str
    usage: int
    max_usage: int
    damage_elemental_type: DamageElementalType
    damage: int
    status_name: str | None = None

    def _create_status(self, match: Any) -> List[CreateObjectAction]:
        """
        Create status
        """
        if self.status_name is not None:  # pragma: no cover
            name = self.status_name
        else:
            name = self.name
        return [
            CreateObjectAction(
                object_name=name,
                object_position=ObjectPosition(
                    player_idx=self.position.player_idx,
                    area=ObjectPositionType.TEAM_STATUS,
                    id=0,
                ),
                object_arguments={},
            )
        ]

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        When created object is self, and not renew, also create teams status
        """
        if event.create_result == "RENEW":
            # renew, do nothing
            return []
        if event.action.object_name == self.name and self.position.check_position_valid(
            event.action.object_position, match, player_idx_same=True, area_same=True
        ):
            # name same, and position same, is self created
            return self._create_status(match)
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        # re-generate status
        return self._create_status(match)

    def _remove(self, match: Any) -> List[RemoveObjectAction]:
        """
        If self should remove, and has generated status, remove together.
        """
        status = match.player_tables[self.position.player_idx].team_status
        target_status = None
        for s in status:
            if s.name == self.name:
                target_status = s
                break
        ret: List[RemoveObjectAction] = [
            RemoveObjectAction(
                object_position=self.position,
            )
        ]
        if target_status is not None:
            ret.append(RemoveObjectAction(object_position=target_status.position))
        return ret



from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....summon.base import AttackerSummonBase
from ....modifiable_values import DamageValue
from ....event import (
    AfterMakeDamageEventArguments,
    ChangeObjectUsageEventArguments,
    RoundEndEventArguments,
    SkillEndEventArguments,
)
from ....status.team_status.base import ShieldTeamStatus, TeamStatusBase
from ....struct import Cost
from ....action import (
    Actions,
    ChangeObjectUsageAction,
    CreateObjectAction,
    DrawCardAction,
    MakeDamageAction,
)
from ....match import Match
from ....consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    IconType,
    ObjectPositionType,
    WeaponType,
)
from ....character.character_base import (
    CharacterBase,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    SkillTalent,
)
from .....utils.desc_registry import DescDictType


class CurtainOfSlumberShield_4_3(ShieldTeamStatus):
    name: Literal["Curtain of Slumber Shield"] = "Curtain of Slumber Shield"
    version: Literal["4.3"] = "4.3"
    usage: int = 2
    max_usage: int = 2


class ShootingStar_4_3(TeamStatusBase):
    """
    Actions related on Shooting Star
    1. Create Shooting Star
    2. After using skill, gain 1 Night Star
    3. Recreate Shooting Star, gain 2 Night Star
    4. When max_usage reached, deal 1 Cryo DMG and clear usage
    5. When deal DMG, if has layla with talent, draw 1 card. (This is
        performed by talent card)
    """

    name: Literal["Shooting Star"] = "Shooting Star"
    version: Literal["4.3"] = "4.3"
    usage: int = 0
    max_usage: int = 4
    icon_type: IconType = IconType.OTHERS

    def renew(self, new_status: "ShootingStar_4_3") -> None:
        """
        When renew, add two usage
        """
        self.usage = min(self.usage + 2, self.max_usage)

    def _make_damage_to_opposite_active_if_max_usage(
        self, match: Match
    ) -> List[MakeDamageAction]:
        """
        Deal 1 Cryo DMG to opposite active and reset usage if max usage reached
        """
        if self.usage < self.max_usage:
            # not reach max usage, return
            return []
        self.usage = 0
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target.position,
                        damage=1,
                        damage_elemental_type=DamageElementalType.CRYO,
                        cost=Cost(),
                    )
                ]
            )
        ]

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectAction, match: Match
    ) -> List[MakeDamageAction]:
        return self._make_damage_to_opposite_active_if_max_usage(match)

    def event_handler_CHANGE_OBJECT_USAGE(
        self, event: ChangeObjectUsageEventArguments, match: Match
    ) -> List[MakeDamageAction]:
        return self._make_damage_to_opposite_active_if_max_usage(match)

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[MakeDamageAction]:
        """
        When self character use skill, gain 1 Night Star and check if need to
        deal damage
        """
        ret: List[MakeDamageAction] = []
        ret += self._make_damage_to_opposite_active_if_max_usage(match)
        if event.action.position.player_idx == self.position.player_idx:
            # self character use skill, gain 1 Night Star
            self.usage = min(self.usage + 1, self.max_usage)
        return ret + self._make_damage_to_opposite_active_if_max_usage(match)


class CelestialDreamsphere_4_3(AttackerSummonBase):
    name: Literal["Celestial Dreamsphere"] = "Celestial Dreamsphere"
    version: Literal["4.3"] = "4.3"
    usage: int = 2
    max_usage: int = 2
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.CRYO

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret: List[MakeDamageAction | ChangeObjectUsageAction] = []
        ret += super().event_handler_ROUND_END(event, match)
        team_status = match.player_tables[self.position.player_idx].team_status
        for status in team_status:
            if status.name == "Shooting Star":
                # Shooting Star exist, gain 1 Night Star
                ret.append(
                    ChangeObjectUsageAction(
                        object_position=status.position, change_usage=1
                    )
                )
                break
        return ret


class NightsOfFormalFocus(ElementalSkillBase):
    name: Literal["Nights of Formal Focus"] = "Nights of Formal Focus"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Generate shield and shooting star
        """
        return [
            self.create_team_status("Curtain of Slumber Shield"),
            self.create_team_status("Shooting Star"),
            self.charge_self(1),
        ]


class DreamOfTheStarStreamShaker(ElementalBurstBase):
    name: Literal["Dream of the Star-Stream Shaker"] = "Dream of the Star-Stream Shaker"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.CRYO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_summon("Celestial Dreamsphere"),
        ]


class LightsRemit_4_3(SkillTalent):
    name: Literal["Light's Remit"]
    version: Literal["4.3"] = "4.3"
    character_name: Literal["Layla"] = "Layla"
    skill: Literal["Nights of Formal Focus"] = "Nights of Formal Focus"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Match
    ) -> List[DrawCardAction]:
        """
        When equipped, self Shooting Star deal damage, draw 1 card
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped, return
            return []
        if (
            self.position.player_idx
            != event.damages[0].final_damage.position.player_idx
        ):
            # not self attack, return
            return []
        obj = match.get_object(event.damages[0].final_damage.position)
        if obj is None:  # pragma: no cover
            # obj is None, return
            return []
        if obj.name != "Shooting Star":
            # not Shooting Star, return
            return []
        # self Shooting Star deal damage, draw 1 card
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ]


class Layla_4_3(CharacterBase):
    name: Literal["Layla"]
    version: Literal["4.3"] = "4.3"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | NightsOfFormalFocus | DreamOfTheStarStreamShaker
    ] = []
    faction: List[FactionType] = [FactionType.SUMERU]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Sword of the Radiant Path",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            NightsOfFormalFocus(),
            DreamOfTheStarStreamShaker(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTER/Layla": {
        "names": {"en-US": "Layla", "zh-CN": "莱依拉"},
        "image_path": "cardface/Char_Avatar_Layla.png",  # noqa: E501
        "id": 1109,
        "descs": {"4.3": {"en-US": "", "zh-CN": ""}},
    },
    "SKILL_Layla_NORMAL_ATTACK/Sword of the Radiant Path": {
        "names": {"en-US": "Sword of the Radiant Path", "zh-CN": "熠辉轨度剑"},
        "descs": {
            "4.3": {"en-US": "Deals 2 Physical DMG.", "zh-CN": "造成2点物理伤害。"}
        },
    },
    "SKILL_Layla_ELEMENTAL_SKILL/Nights of Formal Focus": {
        "names": {"en-US": "Nights of Formal Focus", "zh-CN": "垂裳端凝之夜"},
        "descs": {
            "4.3": {
                "en-US": "Creates Curtain of Slumber Shield and Shooting Star, and collects 1 Night Star for Shooting Star. (If Shooting Star already exists, then collect an extra 2 Night Stars)",  # noqa: E501
                "zh-CN": "生成安眠帷幕护盾和飞星，并为飞星累积1枚「晚星」。（如果飞星已存在，则额外累积2枚「晚星」）",  # noqa: E501
            }
        },
    },
    "TEAM_STATUS/Curtain of Slumber Shield": {
        "names": {"en-US": "Curtain of Slumber Shield", "zh-CN": "安眠帷幕护盾"},
        "descs": {
            "4.3": {
                "en-US": "Provides 2 points of Shield to your active character.",  # noqa: E501
                "zh-CN": "提供2点护盾，保护我方出战角色。",
            }
        },
    },
    "TEAM_STATUS/Shooting Star": {
        "names": {"en-US": "Shooting Star", "zh-CN": "飞星"},
        "descs": {
            "4.3": {
                "en-US": "After one of your characters uses a Skill: Gain 1 Night Star. If possible, consume 4 Night Stars to deal 1 Cryo DMG.",  # noqa: E501
                "zh-CN": "我方角色使用技能后：累积1枚「晚星」。如可能，就消耗4枚「晚星」，造成1点冰元素伤害。",  # noqa: E501
            }
        },
        "image_path": "status/Layla_S.png",
    },
    "SKILL_Layla_ELEMENTAL_BURST/Dream of the Star-Stream Shaker": {
        "names": {"en-US": "Dream of the Star-Stream Shaker", "zh-CN": "星流摇床之梦"},
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Cryo DMG, summons 1 Celestial Dreamsphere.",
                "zh-CN": "造成3点冰元素伤害，召唤饰梦天球。",
            }
        },
    },
    "SUMMON/Celestial Dreamsphere": {
        "names": {"en-US": "Celestial Dreamsphere", "zh-CN": "饰梦天球"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deal 1 Cryo DMG. If Shooting Star is in play, it will gain 1 Night Star.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "结束阶段：造成1点冰元素伤害。如果飞星在场，则使其累积1枚「晚星」。\n可用次数：2",  # noqa: E501
            }
        },
    },
    "TALENT_Layla/Light's Remit": {
        "names": {"en-US": "Light's Remit", "zh-CN": "归芒携信"},
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Layla, equip this card.\nAfter Layla equips this card, immediately use Nights of Formal Focus.\nWhen Layla is on the field and has this card equipped, every time Shooting Star deals DMG, draw 1 card.\n(You must have Layla in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为莱依拉时，装备此牌。\n莱依拉装备此牌后，立刻使用一次垂裳端凝之夜。\n装备有此牌的莱依拉在场时，每当飞星造成伤害，就抓1张牌。\n（牌组中包含莱依拉，才能加入牌组）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Constellation_Layla.png",  # noqa: E501
        "id": 211091,
    },
}


register_class(
    Layla_4_3
    | CurtainOfSlumberShield_4_3
    | ShootingStar_4_3
    | CelestialDreamsphere_4_3
    | LightsRemit_4_3,
    desc,
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import RoundEndEventArguments

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue

from ...action import ActionTypes, Actions, ChangeObjectUsageAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE,
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalNormalAttackBase,
    ElementalSkillBase,
    CharacterBase,
    SkillTalent,
)


# Summons


class BakeKurage_3_5(AttackerSummonBase):
    name: Literal["Bake-Kurage"] = "Bake-Kurage"
    version: Literal["3.5"] = "3.5"
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        active_character = match.player_tables[
            self.position.player_idx
        ].get_active_character()
        make_damage_action = ret[0]
        assert make_damage_action.type == ActionTypes.MAKE_DAMAGE
        make_damage_action.damage_value_list.append(
            DamageValue(
                position=self.position,
                damage_type=DamageType.HEAL,
                target_position=active_character.position,
                damage=-1,
                damage_elemental_type=DamageElementalType.HEAL,
                cost=Cost(),
            )
        )
        # if our Kokomi with talent and Ceremonial Garment, increase damage
        characters = match.player_tables[self.position.player_idx].characters
        is_kokomi_talent_status: bool = False
        for c in characters:
            if c.name == "Sangonomiya Kokomi" and c.talent is not None:
                # our Kokomi with talent
                status_found = False
                for s in c.status:
                    if s.name == "Ceremonial Garment":
                        # Ceremonial Garment exists
                        status_found = True
                        break
                if status_found:
                    # Ceremonial Garment exists
                    is_kokomi_talent_status = True
                    break
        if is_kokomi_talent_status:
            # increase damage
            make_damage_action.damage_value_list[0].damage += 1
        return ret


# Skills


class KuragesOath(ElementalSkillBase):
    name: Literal["Kurage's Oath"] = "Kurage's Oath"
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        application and create object
        """
        ret = [
            self.element_application_self(match, DamageElementalType.HYDRO),
            self.create_summon("Bake-Kurage"),
            self.charge_self(1),
        ]
        return ret


class NereidsAscension(ElementalBurstBase):
    name: Literal["Nereid's Ascension"] = "Nereid's Ascension"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        damage, heal and create character status
        """
        ret = super().get_actions(match)
        damage_action = ret[1]
        assert damage_action.type == ActionTypes.MAKE_DAMAGE
        characters = match.player_tables[self.position.player_idx].characters
        for character in characters:
            if character.is_alive:
                damage_action.damage_value_list.append(
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.HEAL,
                        target_position=character.position,
                        damage=-1,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                )
        ret.append(self.create_character_status("Ceremonial Garment"))
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        summons = match.player_tables[self.position.player_idx].summons
        kurage_found = None
        for s in summons:
            if s.name == "Bake-Kurage":
                kurage_found = s
                break
        talent = character.talent
        if talent is not None:
            if talent.version == "3.5":
                # if has 3.5 talent, re-create Bake-Kurage
                if kurage_found:
                    ret.append(self.create_summon("Bake-Kurage"))
            elif talent.version == "4.2":
                # with 4.2 talent
                if kurage_found is not None:
                    # found kurage, increase usage
                    ret.append(
                        ChangeObjectUsageAction(
                            object_position=kurage_found.position, change_usage=1
                        )
                    )
                else:
                    # otherwise, create kurage with 1 usage
                    ret.append(
                        self.create_summon(
                            "Bake-Kurage",
                            {
                                "usage": 1,
                                "max_usage": 1,
                            },
                        )
                    )
            else:
                raise NotImplementedError
        return ret


# Talents


class TamakushiCasket_4_2(SkillTalent):
    name: Literal["Tamakushi Casket"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Sangonomiya Kokomi"] = "Sangonomiya Kokomi"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.HYDRO, elemental_dice_number=3, charge=2
    )
    skill: Literal["Nereid's Ascension"] = "Nereid's Ascension"


# character base


class SangonomiyaKokomi_3_6(CharacterBase):
    name: Literal["Sangonomiya Kokomi"]
    version: Literal["3.6"] = "3.6"
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[ElementalNormalAttackBase | KuragesOath | NereidsAscension] = []
    faction: List[FactionType] = [FactionType.INAZUMA]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name="The Shape of Water",
                damage_type=ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost=ElementalNormalAttackBase.get_cost(self.element),
            ),
            KuragesOath(),
            NereidsAscension(),
        ]


register_class(SangonomiyaKokomi_3_6 | TamakushiCasket_4_2 | BakeKurage_3_5)



from lpsim.server.interaction import SwitchCardResponse
from tests.utils_for_test import make_respond
from lpsim.agents.interaction_agent import InteractionAgent
from lpsim.agents.nothing_agent import NothingAgent
from lpsim.server.match import Match
from lpsim.server.deck import Deck


def test_draw_card():
    """
    when all card same, should raise no error and hand number right
    """
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida*3
        Strategize*30
        """
    )
    match = Match()
    match.set_deck([deck, deck])
    match.config.max_same_card_number = 30
    match.config.random_first_player = False
    match.start()
    match.step()
    agent_0 = NothingAgent(player_idx=0)
    agent_1 = InteractionAgent(
        player_idx=1,
        commands=[
            "sw_card 1 2 3 4",
        ],
        only_use_command=True,
    )
    while True:
        if match.need_respond(0):
            make_respond(agent_0, match)
        elif match.need_respond(1):
            make_respond(agent_1, match)
        else:
            raise AssertionError("No need respond.")
        if len(agent_1.commands) == 0:
            break
    assert len(agent_1.commands) == 0
    assert len(match.player_tables[1].hands) == 5

    """
    when two types of cards, should become another type in all hand
    """
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Nahida*3
        Strategize*15
        Rana*15
        """
    )
    for _ in range(100):
        match = Match()
        match.set_deck([deck, deck])
        match.config.max_same_card_number = 30
        match.config.random_first_player = False
        match.start()
        match.step()
        agent_0 = NothingAgent(player_idx=0)
        while True:
            if match.need_respond(0):
                make_respond(agent_0, match)
            elif match.need_respond(1):
                assert len(match.requests) == 1
                assert match.requests[0].name == "SwitchCardRequest"
                req = match.requests[0]
                idxs = []
                for id, name in enumerate(req.card_names):
                    if name == "Strategize":
                        idxs.append(id)
                resp: SwitchCardResponse = SwitchCardResponse(
                    request=req,
                    card_idxs=idxs,
                )
                match.respond(resp)
                match.step()
                break
            else:
                raise AssertionError("No need respond.")
        assert len(match.player_tables[1].hands) == 5
        for card in match.player_tables[1].hands:
            assert card.name == "Rana"


if __name__ == "__main__":
    test_draw_card()



from lpsim.server.event_handler import OmnipotentGuideEventHandler_3_3
from lpsim.network import HTTPServer
from lpsim.server.match import MatchConfig, Match
from tests.utils_for_test import get_random_state
import logging


logging.basicConfig(level=logging.INFO)


if __name__ == "__main__":
    deck_str_1 = """
        default_version:4.4
        character:AnemoMobMage
        character:CryoMob
        character:PyroMobMage
        character:Sangonomiya Kokomi
        Mushroom Pizza*10
        Matsutake Meat Rolls*10
        Silver and Melus*10
    """
    deck_str_2 = """
    character:Nahida@3.7
    character:Rhodeia of Loch@3.3
    character:Fischl@3.3
    Gambler's Earrings@3.8
    Paimon@3.3
    Chef Mao@4.1
    Dunyarzad@4.1
    The Bestest Travel Companion!@3.3
    Paimon@3.3
    Liben@3.3
    Liben@3.3
    Send Off@3.7
    Teyvat Fried Egg@4.1
    Dunyarzad@4.1
    Sweet Madame@3.3
    I Haven't Lost Yet!@4.0
    Send Off@3.7
    Lotus Flower Crisp@3.3
    Magic Guide@3.3*15
    """
    server = HTTPServer(
        decks=[deck_str_1, deck_str_1],
        match_config=MatchConfig(
            check_deck_restriction=False,
            card_number=None,
            max_same_card_number=None,
            character_number=None,
            max_round_number=999,
            random_first_player=False,
            # max_hand_size = 999,
            # recreate_mode = True,
            history_level=10,
            # make_skill_prediction = True,
            # random_object_information = {
            #     'rhodeia': [
            #         'squirrel', 'raptor', 'frog', 'squirrel', 'raptor',
            #         'frog', 'frog', 'squirrel', 'squirrel'
            #     ],
            # }
        ),
    )

    # # modify hp
    # for deck in server.decks:
    #     for character in deck.characters:
    #         character.hp = 40
    #         character.max_hp = 40

    # fix random seed
    match = Match(random_state=get_random_state())
    match.set_deck(server.decks)
    match.config = server.match.config

    # rich mode
    match.event_handlers.append(OmnipotentGuideEventHandler_3_3())
    match.config.initial_dice_number = 16

    server.match = match
    server.match_random_state = get_random_state()
    server.match.start()
    server.match._save_history()
    server.match.step()
    server.run()
